#!/usr/bin/env python3 -B
# coding=utf-8

"""
Insyde IFD Extract
Insyde iFlash/iFdPacker Extractor
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import ctypes
import os
import re

from typing import Any, Final

from biosutilities.common.compression import is_szip_supported, szip_decompress
from biosutilities.common.paths import (delete_file, extract_folder, is_file_read, path_files,
                                        make_dirs, path_name, safe_name)
from biosutilities.common.patterns import PAT_INSYDE_IFL, PAT_INSYDE_SFX
from biosutilities.common.structs import CHAR, ctypes_struct, UINT32
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility


class IflashHeader(ctypes.LittleEndianStructure):
    """ Insyde iFlash Header """

    _pack_ = 1
    _fields_ = [
        ('Signature',       CHAR * 8),      # 0x00 $_IFLASH
        ('ImageTag',        CHAR * 8),      # 0x08
        ('TotalSize',       UINT32),        # 0x10 from header end
        ('ImageSize',       UINT32)         # 0x14 from header end
        # 0x18
    ]

    def _get_padd_len(self) -> int:
        return self.TotalSize - self.ImageSize

    def get_image_tag(self) -> str:
        """ Get Insyde iFlash image tag """

        return self.ImageTag.decode('utf-8', 'ignore').strip('_')

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Signature :', self.Signature.decode('utf-8')], padding=padding, new_line=False)
        printer(message=['Image Name:', self.get_image_tag()], padding=padding, new_line=False)
        printer(message=['Image Size:', f'0x{self.ImageSize:X}'], padding=padding, new_line=False)
        printer(message=['Total Size:', f'0x{self.TotalSize:X}'], padding=padding, new_line=False)
        printer(message=['Padd Size :', f'0x{self._get_padd_len():X}'], padding=padding, new_line=False)


class InsydeIfdExtract(BIOSUtility):
    """ Insyde iFlash/iFdPacker Extractor """

    TITLE: str = 'Insyde iFlash/iFdPacker Extractor'

    # Insyde iFdPacker known 7-Zip SFX Password
    INS_SFX_PWD: Final[str] = 'Y`t~i!L@i#t$U%h^s7A*l(f)E-d=y+S_n?i'

    # Insyde iFlash known Image Names
    INS_IFL_IMG: Final[dict[str, list[str]]] = {
        'BIOSCER': ['Certificate', 'bin'],
        'BIOSCR2': ['Certificate 2nd', 'bin'],
        'BIOSIMG': ['BIOS-UEFI', 'bin'],
        'DRV_IMG': ['isflash', 'efi'],
        'EC_IMG': ['Embedded Controller', 'bin'],
        'INI_IMG': ['platform', 'ini'],
        'IOM_IMG': ['IO Manageability', 'bin'],
        'ISH_IMG': ['Integrated Sensor Hub', 'bin'],
        'ME_IMG': ['Management Engine', 'bin'],
        'OEM_ID': ['OEM Identifier', 'bin'],
        'PDT_IMG': ['Platform Descriptor Table', 'bin'],
        'TBT_IMG': ['Integrated Thunderbolt', 'bin']
    }

    # Get common ctypes Structure Sizes
    INS_IFL_LEN: Final[int] = ctypes.sizeof(IflashHeader)

    def check_format(self) -> bool:
        """ Check if input is Insyde iFlash/iFdPacker Update image """

        if bool(self._insyde_iflash_detect(input_buffer=self.input_buffer)):
            return True

        if bool(PAT_INSYDE_SFX.search(self.input_buffer)):
            return True

        return False

    def parse_format(self) -> bool:
        """ Parse & Extract Insyde iFlash/iFdPacker Update images """

        iflash_code: int = self._insyde_iflash_extract(input_buffer=self.input_buffer, extract_path=self.extract_path,
                                                       padding=self.padding)

        ifdpack_path: str = os.path.join(self.extract_path, 'Insyde iFdPacker SFX')

        ifdpack_code: int = self._insyde_packer_extract(input_buffer=self.input_buffer, extract_path=ifdpack_path,
                                                        padding=self.padding)

        return (iflash_code and ifdpack_code) == 0

    def _insyde_iflash_detect(self, input_buffer: bytes) -> list:
        """ Detect Insyde iFlash Update image """

        iflash_match_all: list = []
        iflash_match_nan: list = [0x0, 0xFFFFFFFF]

        for iflash_match in PAT_INSYDE_IFL.finditer(input_buffer):
            ifl_bgn: int = iflash_match.start()

            if len(input_buffer[ifl_bgn:]) <= self.INS_IFL_LEN:
                continue

            ifl_hdr: Any = ctypes_struct(buffer=input_buffer, start_offset=ifl_bgn, class_object=IflashHeader)

            if ifl_hdr.TotalSize in iflash_match_nan \
                    or ifl_hdr.ImageSize in iflash_match_nan \
                    or ifl_hdr.TotalSize < ifl_hdr.ImageSize \
                    or ifl_bgn + self.INS_IFL_LEN + ifl_hdr.TotalSize > len(input_buffer):
                continue

            iflash_match_all.append([ifl_bgn, ifl_hdr])

        return iflash_match_all

    def _insyde_iflash_extract(self, input_buffer: bytes, extract_path: str, padding: int = 0) -> int:
        """ Extract Insyde iFlash Update image """

        insyde_iflash_all: list = self._insyde_iflash_detect(input_buffer=input_buffer)

        if not insyde_iflash_all:
            return 127

        printer(message='Detected Insyde iFlash Update image!', padding=padding)

        make_dirs(in_path=extract_path)

        exit_codes: list = []

        for insyde_iflash in insyde_iflash_all:
            exit_code: int = 0

            ifl_bgn, ifl_hdr = insyde_iflash

            img_bgn: int = ifl_bgn + self.INS_IFL_LEN
            img_end: int = img_bgn + ifl_hdr.ImageSize
            img_bin: bytes = input_buffer[img_bgn:img_end]

            if len(img_bin) != ifl_hdr.ImageSize:
                exit_code = 1

            img_val: list = [ifl_hdr.get_image_tag(), 'bin']
            img_tag, img_ext = self.INS_IFL_IMG.get(img_val[0], img_val)

            img_name: str = f'{img_tag} [0x{img_bgn:08X}-0x{img_end:08X}]'

            printer(message=f'{img_name}\n', padding=padding + 4)

            ifl_hdr.struct_print(padding=padding + 8)

            if img_val == [img_tag, img_ext]:
                printer(message=f'Note: Detected new Insyde iFlash tag {img_tag}!', padding=padding + 12)

            out_name: str = f'{img_name}.{img_ext}'

            out_path: str = os.path.join(extract_path, safe_name(in_name=out_name))

            with open(out_path, 'wb') as out_image:
                out_image.write(img_bin)

            printer(message=f'Successful Insyde iFlash > {img_tag} extraction!', padding=padding + 12)

            exit_codes.append(exit_code)

        return sum(exit_codes)

    def _insyde_packer_extract(self, input_buffer: bytes, extract_path: str, padding: int = 0) -> int:
        """ Extract Insyde iFdPacker 7-Zip SFX 7z Update image """

        match_sfx: re.Match[bytes] | None = PAT_INSYDE_SFX.search(input_buffer)

        if not match_sfx:
            return 127

        printer(message='Detected Insyde iFdPacker Update image!', padding=padding)

        make_dirs(in_path=extract_path, delete=True)

        sfx_buffer: bytearray = bytearray(input_buffer[match_sfx.end() - 0x5:])

        if sfx_buffer[:0x5] == b'\x6E\xF4\x79\x5F\x4E':
            printer(message='Detected Insyde iFdPacker > 7-Zip SFX > Obfuscation!', padding=padding + 4)

            for index, byte in enumerate(sfx_buffer):
                sfx_buffer[index] = byte // 2 + (128 if byte % 2 else 0)

            printer(message='Removed Insyde iFdPacker > 7-Zip SFX > Obfuscation!', padding=padding + 8)

        printer(message='Extracting Insyde iFdPacker > 7-Zip SFX archive...', padding=padding + 4)

        if bytes(self.INS_SFX_PWD, 'utf-16le') in input_buffer[:match_sfx.start()]:
            printer(message='Detected Insyde iFdPacker > 7-Zip SFX > Password!', padding=padding + 8)

            printer(message=self.INS_SFX_PWD, padding=padding + 12)

        sfx_path: str = os.path.join(extract_path, 'Insyde_iFdPacker_SFX.7z')

        with open(sfx_path, 'wb') as sfx_file_object:
            sfx_file_object.write(sfx_buffer)

        if is_szip_supported(in_path=sfx_path, args=[f'-p{self.INS_SFX_PWD}']):
            if szip_decompress(in_path=sfx_path, out_path=extract_path, in_name='Insyde iFdPacker > 7-Zip SFX',
                               padding=padding + 8, args=[f'-p{self.INS_SFX_PWD}'], check=True):
                delete_file(in_path=sfx_path)
            else:
                return 125
        else:
            return 126

        exit_codes: list[int] = []

        for sfx_file in path_files(in_path=extract_path):
            if is_file_read(in_path=sfx_file):
                insyde_ifd_extract: InsydeIfdExtract = InsydeIfdExtract(
                    input_object=sfx_file, extract_path=extract_folder(sfx_file), padding=padding + 16)

                if insyde_ifd_extract.check_format():
                    printer(message=path_name(in_path=sfx_file), padding=padding + 12)

                    ifd_status: int = insyde_ifd_extract.parse_format()

                    exit_codes.append(0 if ifd_status else 1)

        return sum(exit_codes)



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Toshiba COM Extract
Toshiba BIOS COM Extractor
Copyright (C) 2018-2024 Plato Mavropoulos
"""

import os
import subprocess

from biosutilities.common.externals import comextract_path
from biosutilities.common.paths import delete_file, is_file_read, make_dirs, path_stem
from biosutilities.common.patterns import PAT_TOSHIBA_COM
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility


class ToshibaComExtract(BIOSUtility):
    """ Toshiba BIOS COM Extractor """

    TITLE: str = 'Toshiba BIOS COM Extractor'

    def check_format(self) -> bool:
        """ Check if input is Toshiba BIOS COM image """

        if isinstance(self.input_object, str) and is_file_read(in_path=self.input_object):
            with open(self.input_object, 'rb') as input_object:
                check_buffer: bytes = input_object.read(0x100)
        else:
            check_buffer = self.input_buffer[:0x100]

        return bool(PAT_TOSHIBA_COM.search(check_buffer, 0, 0x100))

    def parse_format(self) -> bool:
        """ Parse & Extract Toshiba BIOS COM image """

        make_dirs(in_path=self.extract_path)

        if isinstance(self.input_object, str) and is_file_read(in_path=self.input_object):
            input_path: str = self.input_object
        else:
            input_path = os.path.join(self.extract_path, 'toshiba_bios.com')

            with open(input_path, 'wb') as input_buffer:
                input_buffer.write(self.input_buffer)

        output_path: str = os.path.join(self.extract_path, f'{path_stem(in_path=input_path)}_extracted.bin')

        comextract_res: subprocess.CompletedProcess[bytes] = subprocess.run(
            [comextract_path(), input_path, output_path], check=False, stdout=subprocess.DEVNULL)

        if input_path != self.input_object:
            delete_file(in_path=input_path)

        if comextract_res.returncode == 0 and is_file_read(in_path=output_path):
            printer(message='Successful extraction via ToshibaComExtractor!', padding=self.padding)

            return True

        return False



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import os
import re
import shutil
import stat
import sys

from pathlib import Path, PurePath
from typing import Callable, Final

from biosutilities.common.system import system_platform
from biosutilities.common.texts import to_string

MAX_WIN_COMP_LEN: Final[int] = 255


def safe_name(in_name: str) -> str:
    """
    Fix illegal/reserved Windows characters
    Can also be used to nuke dangerous paths
    """

    name_repr: str = repr(in_name).strip("'")

    return re.sub(r'[\\/:"*?<>|]+', '_', name_repr)


def safe_path(base_path: str, user_paths: str | list | tuple) -> str:
    """ Check and attempt to fix illegal/unsafe OS path traversals """

    # Convert base path to absolute path
    base_path = real_path(in_path=base_path)

    # Merge user path(s) to string with OS separators
    user_path: str = to_string(in_object=user_paths, sep_char=os.sep)

    # Create target path from base + requested user path
    target_path: str = norm_path(base_path=base_path, user_path=user_path)

    # Check if target path is OS illegal/unsafe
    if is_safe_path(base_path=base_path, target_path=target_path):
        return target_path

    # Re-create target path from base + leveled/safe illegal "path" (now file)
    nuked_path: str = norm_path(base_path=base_path, user_path=safe_name(in_name=user_path))

    # Check if illegal path leveling worked
    if is_safe_path(base_path=base_path, target_path=nuked_path):
        return nuked_path

    # Still illegal, raise exception to halt execution
    raise OSError(f'Encountered illegal path traversal: {user_path}')


def is_safe_path(base_path: str, target_path: str) -> bool:
    """ Check for illegal/unsafe OS path traversal """

    base_path = real_path(in_path=base_path)

    target_path = real_path(in_path=target_path)

    common_path: str = os.path.commonpath((base_path, target_path))

    return base_path == common_path


def norm_path(base_path: str, user_path: str) -> str:
    """ Create normalized base path + OS separator + user path """

    return os.path.normpath(base_path + os.sep + user_path)


def real_path(in_path: str) -> str:
    """ Get absolute path, resolving any symlinks """

    return os.path.realpath(in_path)


def agnostic_path(in_path: str) -> PurePath:
    """ Get Windows/Posix OS-agnostic path """

    return PurePath(in_path.replace('\\', os.sep))


def path_parent(in_path: str) -> Path:
    """ Get absolute parent of path """

    return Path(in_path).parent.absolute()


def path_name(in_path: str, limit: bool = False) -> str:
    """ Get final path component, with suffix """

    comp_name: str = PurePath(in_path).name

    is_win: bool = system_platform()[1]

    if limit and is_win:
        comp_name = comp_name[:MAX_WIN_COMP_LEN - len(extract_suffix())]

    return comp_name


def path_stem(in_path: str) -> str:
    """ Get final path component, w/o suffix """

    return PurePath(in_path).stem


def path_size(in_path: str) -> int:
    """ Get path size (bytes) """

    return os.stat(in_path).st_size


def path_suffixes(in_path: str) -> list[str]:
    """ Get list of path file extensions """

    return PurePath(in_path).suffixes or ['']


def make_dirs(in_path: str, parents: bool = True, exist_ok: bool = True, delete: bool = False):
    """ Create folder(s), controlling parents, existence and prior deletion """

    if delete:
        delete_dirs(in_path=in_path)

    Path.mkdir(Path(in_path), parents=parents, exist_ok=exist_ok)


def delete_dirs(in_path: str) -> None:
    """ Delete folder(s), if present """

    if is_dir(in_path=in_path):
        shutil.rmtree(in_path, onerror=clear_readonly_callback)  # pylint: disable=deprecated-argument


def delete_file(in_path: str) -> None:
    """ Delete file, if present """

    if is_file(in_path=in_path):
        clear_readonly(in_path=in_path)

        os.remove(in_path)


def rename_file(in_path: str, in_dest: str) -> None:
    """ Rename file with path or name destination, if present """

    if is_file(in_path=in_path):
        clear_readonly(in_path=in_path)

        if is_file(in_path=in_dest, allow_broken_links=True):
            clear_readonly(in_path=in_dest)

            out_path: str = in_dest
        else:
            out_path = os.path.join(path_parent(in_path=in_path), in_dest)

        os.replace(in_path, out_path)


def copy_file(in_path: str, out_path: str, metadata: bool = False) -> None:
    """ Copy file to path with or w/o metadata """

    if metadata:
        shutil.copy2(in_path, out_path)
    else:
        shutil.copy(in_path, out_path)


def clear_readonly(in_path: str) -> None:
    """ Clear read-only file attribute """

    os.chmod(in_path, stat.S_IWRITE)


def clear_readonly_callback(in_func: Callable, in_path: str, _) -> None:
    """ Clear read-only file attribute (on shutil.rmtree error) """

    clear_readonly(in_path=in_path)

    in_func(in_path)


def path_files(in_path: str, follow_links: bool = False, root_only: bool = False) -> list[str]:
    """ Walk path to get all files """

    file_paths: list[str] = []

    for root_path, _, file_names in os.walk(in_path, followlinks=follow_links):
        for file_name in file_names:
            file_path: str = os.path.abspath(os.path.join(root_path, file_name))

            if is_file(in_path=file_path):
                file_paths.append(file_path)

        if root_only:
            break

    return file_paths


def is_dir(in_path: str) -> bool:
    """ Check if path is a directory """

    return Path(in_path).is_dir()


def is_file(in_path: str, allow_broken_links: bool = False) -> bool:
    """ Check if path is a regural file or symlink (valid or broken) """

    in_path_abs: str = os.path.abspath(in_path)

    if os.path.lexists(in_path_abs):
        if not is_dir(in_path=in_path_abs):
            if allow_broken_links:
                return os.path.isfile(in_path_abs) or os.path.islink(in_path_abs)

            return os.path.isfile(in_path_abs)

    return False


def is_access(in_path: str, access_mode: int = os.R_OK, follow_links: bool = False) -> bool:
    """ Check if path is accessible """

    if not follow_links and os.access not in os.supports_follow_symlinks:
        follow_links = True

    return os.access(in_path, access_mode, follow_symlinks=follow_links)


def is_file_read(in_path: str) -> bool:
    """ Check if path is a readable file """

    return isinstance(in_path, str) and is_file(in_path=in_path) and is_access(in_path=in_path)


def is_dir_read(in_path: str) -> bool:
    """ Check if path is a readable directory """

    return isinstance(in_path, str) and is_dir(in_path=in_path) and is_access(in_path=in_path)


def is_empty_dir(in_path: str, follow_links: bool = False) -> bool:
    """ Check if directory is empty (file-wise) """

    for _, _, filenames in os.walk(in_path, followlinks=follow_links):
        if filenames:
            return False

    return True


def extract_suffix() -> str:
    """ Set utility extraction stem """

    return '_extracted'


def extract_folder(in_path: str, suffix: str = extract_suffix()) -> str:
    """ Get utility extraction directory """

    return f'{in_path}{suffix}'


def project_root() -> str:
    """ Get project root directory """

    return real_path(in_path=str(Path(__file__).parent.parent))


def runtime_root() -> str:
    """ Get runtime root directory """

    if getattr(sys, 'frozen', False):
        root: str = str(Path(sys.executable).parent)
    else:
        root = project_root()

    return real_path(in_path=root)



#!/usr/bin/env python3 -B
# coding=utf-8

"""
AMI PFAT Extract
AMI BIOS Guard Extractor
Copyright (C) 2018-2024 Plato Mavropoulos
"""

import ctypes
import os
import re
import struct

from typing import Any, Final, Type

from biosutilities.common.externals import big_script_tool
from biosutilities.common.paths import extract_suffix, extract_folder, make_dirs, path_name, safe_name
from biosutilities.common.patterns import PAT_AMI_PFAT
from biosutilities.common.structs import CHAR, ctypes_struct, UINT8, UINT16, UINT32
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility
from biosutilities.common.texts import bytes_to_hex, file_to_bytes, to_ordinal


class AmiBiosGuardHeader(ctypes.LittleEndianStructure):
    """ AMI BIOS Guard Header """

    _pack_ = 1
    _fields_ = [
        ('Size',            UINT32),        # 0x00 Header + Entries
        ('Checksum',        UINT32),        # 0x04 ?
        ('Tag',             CHAR * 8),      # 0x04 _AMIPFAT
        ('Flags',           UINT8)          # 0x10 ?
        # 0x11
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Size    :', f'0x{self.Size:X}'], padding=padding, new_line=False)
        printer(message=['Checksum:', f'0x{self.Checksum:04X}'], padding=padding, new_line=False)
        printer(message=['Tag     :', self.Tag.decode('utf-8')], padding=padding, new_line=False)
        printer(message=['Flags   :', f'0x{self.Flags:02X}'], padding=padding, new_line=False)


class IntelBiosGuardHeader(ctypes.LittleEndianStructure):
    """ Intel BIOS Guard Header """

    _pack_ = 1
    _fields_ = [
        ('BGVerMajor',      UINT16),        # 0x00
        ('BGVerMinor',      UINT16),        # 0x02
        ('PlatformID',      UINT8 * 16),    # 0x04
        ('Attributes',      UINT32),        # 0x14
        ('ScriptVerMajor',  UINT16),        # 0x16
        ('ScriptVerMinor',  UINT16),        # 0x18
        ('ScriptSize',      UINT32),        # 0x1C
        ('DataSize',        UINT32),        # 0x20
        ('BIOSSVN',         UINT32),        # 0x24
        ('ECSVN',           UINT32),        # 0x28
        ('VendorInfo',      UINT32)         # 0x2C
        # 0x30
    ]

    def get_platform_id(self) -> str:
        """ Get Intel BIOS Guard Platform ID """

        id_byte: bytes = bytes(self.PlatformID)

        id_text: str = re.sub(r'[\n\t\r\x00 ]', '', id_byte.decode('utf-8', 'ignore'))

        id_hexs: str = f'{int.from_bytes(id_byte, byteorder="big"):0{0x10 * 2}X}'

        id_guid: str = f'{{{id_hexs[:8]}-{id_hexs[8:12]}-{id_hexs[12:16]}-{id_hexs[16:20]}-{id_hexs[20:]}}}'

        return f'{id_text} {id_guid}'

    def get_hdr_marker(self) -> bytes:
        """ Get Intel BIOS Guard Header Marker """

        return struct.pack('<HH16B', self.BGVerMajor, self.BGVerMinor, *self.PlatformID)

    def get_flags(self) -> tuple:
        """ Get Intel BIOS Guard Header Attributes """

        attr: IntelBiosGuardHeaderGetAttributes = IntelBiosGuardHeaderGetAttributes()

        attr.asbytes = self.Attributes  # pylint: disable=attribute-defined-outside-init

        return attr.b.SFAM, attr.b.ProtectEC, attr.b.GFXMitDis, attr.b.FTU, attr.b.Reserved

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        no_yes: dict[int, str] = {0: 'No', 1: 'Yes'}

        sfam, ec_opc, gfx_dis, ft_upd, attr_res = self.get_flags()

        bg_version: str = f'{self.BGVerMajor}.{self.BGVerMinor}'
        script_version: str = f'{self.ScriptVerMajor}.{self.ScriptVerMinor}'

        printer(message=['BIOS Guard Version          :', bg_version], padding=padding, new_line=False)
        printer(message=['Platform Identity           :', self.get_platform_id()], padding=padding, new_line=False)
        printer(message=['Signed Flash Address Map    :', no_yes[sfam]], padding=padding, new_line=False)
        printer(message=['Protected EC OpCodes        :', no_yes[ec_opc]], padding=padding, new_line=False)
        printer(message=['Graphics Security Disable   :', no_yes[gfx_dis]], padding=padding, new_line=False)
        printer(message=['Fault Tolerant Update       :', no_yes[ft_upd]], padding=padding, new_line=False)
        printer(message=['Attributes Reserved         :', f'0x{attr_res:X}'], padding=padding, new_line=False)
        printer(message=['Script Version              :', script_version], padding=padding, new_line=False)
        printer(message=['Script Size                 :', f'0x{self.ScriptSize:X}'], padding=padding, new_line=False)
        printer(message=['Data Size                   :', f'0x{self.DataSize:X}'], padding=padding, new_line=False)
        printer(message=['BIOS Security Version Number:', f'0x{self.BIOSSVN:X}'], padding=padding, new_line=False)
        printer(message=['EC Security Version Number  :', f'0x{self.ECSVN:X}'], padding=padding, new_line=False)
        printer(message=['Vendor Information          :', f'0x{self.VendorInfo:X}'], padding=padding, new_line=False)


class IntelBiosGuardHeaderAttributes(ctypes.LittleEndianStructure):
    """ Intel BIOS Guard Header Attributes """

    _pack_ = 1
    _fields_ = [
        ('SFAM',            UINT32,     1),     # Signed Flash Address Map
        ('ProtectEC',       UINT32,     1),     # Protected EC OpCodes
        ('GFXMitDis',       UINT32,     1),     # GFX Security Disable
        ('FTU',             UINT32,     1),     # Fault Tolerant Update
        ('Reserved',        UINT32,     28)     # Reserved/Unknown
    ]


class IntelBiosGuardHeaderGetAttributes(ctypes.Union):
    """ Intel BIOS Guard Header Attributes Getter """

    _pack_ = 1
    _fields_ = [
        ('b',               IntelBiosGuardHeaderAttributes),
        ('asbytes',         UINT32)
    ]


class IntelBiosGuardSignatureHeader(ctypes.LittleEndianStructure):
    """ Intel BIOS Guard Signature Header """

    _pack_ = 1
    _fields_ = [
        ('Unknown0',        UINT32),        # 0x000
        ('Unknown1',        UINT32)         # 0x004
        # 0x8
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Unknown 0:', f'0x{self.Unknown0:X}'], padding=padding, new_line=False)
        printer(message=['Unknown 1:', f'0x{self.Unknown1:X}'], padding=padding, new_line=False)


class IntelBiosGuardSignatureRsa2k(ctypes.LittleEndianStructure):
    """ Intel BIOS Guard Signature Block 2048-bit """

    _pack_ = 1
    _fields_ = [
        ('Modulus',         UINT8 * 256),   # 0x000
        ('Exponent',        UINT32),        # 0x100
        ('Signature',       UINT8 * 256)    # 0x104
        # 0x204
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        modulus: str = f'{bytes_to_hex(in_buffer=self.Modulus, order="little", data_len=0x100, slice_len=32)} [...]'
        exponent: str = f'0x{self.Exponent:X}'
        signature: str = f'{bytes_to_hex(in_buffer=self.Signature, order="little", data_len=0x100, slice_len=32)} [...]'

        printer(message=['Modulus  :', modulus], padding=padding, new_line=False)
        printer(message=['Exponent :', exponent], padding=padding, new_line=False)
        printer(message=['Signature:', signature], padding=padding, new_line=False)


class IntelBiosGuardSignatureRsa3k(ctypes.LittleEndianStructure):
    """ Intel BIOS Guard Signature Block 3072-bit """

    _pack_ = 1
    _fields_ = [
        ('Modulus',         UINT8 * 384),   # 0x000
        ('Exponent',        UINT32),        # 0x180
        ('Signature',       UINT8 * 384)    # 0x184
        # 0x304
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        modulus: str = f'{int.from_bytes(self.Modulus, byteorder="little"):0{0x180 * 2}X}'[:64]
        exponent: str = f'0x{self.Exponent:X}'
        signature: str = f'{int.from_bytes(self.Signature, byteorder="little"):0{0x180 * 2}X}'[:64]

        printer(message=['Modulus  :', modulus], padding=padding, new_line=False)
        printer(message=['Exponent :', exponent], padding=padding, new_line=False)
        printer(message=['Signature:', signature], padding=padding, new_line=False)


class AmiPfatExtract(BIOSUtility):
    """ AMI BIOS Guard Extractor """

    TITLE: str = 'AMI BIOS Guard Extractor'

    PFAT_AMI_HDR_LEN: Final[int] = ctypes.sizeof(AmiBiosGuardHeader)
    PFAT_INT_HDR_LEN: Final[int] = ctypes.sizeof(IntelBiosGuardHeader)
    PFAT_INT_SIG_HDR_LEN: Final[int] = ctypes.sizeof(IntelBiosGuardSignatureHeader)
    PFAT_INT_SIG_R2K_LEN: Final[int] = ctypes.sizeof(IntelBiosGuardSignatureRsa2k)
    PFAT_INT_SIG_R3K_LEN: Final[int] = ctypes.sizeof(IntelBiosGuardSignatureRsa3k)
    PFAT_INT_SIG_MAX_LEN: Final[int] = PFAT_INT_SIG_HDR_LEN + PFAT_INT_SIG_R3K_LEN

    def check_format(self) -> bool:
        """ Check if input is AMI BIOS Guard """

        return bool(self._get_ami_pfat(input_object=self.input_buffer))

    def parse_format(self) -> bool:
        """ Process and store AMI BIOS Guard output file """

        pfat_buffer: bytes = self._get_ami_pfat(input_object=self.input_buffer)

        file_path: str = ''

        all_blocks_dict: dict = {}

        bg_sign_len: int = 0

        extract_name: str = path_name(in_path=self.extract_path).removesuffix(extract_suffix())

        make_dirs(in_path=self.extract_path)

        block_all, block_off, file_count = self._parse_pfat_hdr(buffer=pfat_buffer, padding=self.padding)

        for block in block_all:
            file_desc, file_name, _, _, _, file_index, block_index, block_count = block

            if block_index == 0:
                printer(message=file_desc, padding=self.padding + 4)

                file_path = os.path.join(self.extract_path, self._get_file_name(index=file_index + 1, name=file_name))

                all_blocks_dict[file_index] = b''

            block_status: str = f'{block_index + 1}/{block_count}'

            bg_hdr: Any = ctypes_struct(buffer=pfat_buffer, start_offset=block_off, class_object=IntelBiosGuardHeader)

            printer(message=f'Intel BIOS Guard {block_status} Header:\n', padding=self.padding + 8)

            bg_hdr.struct_print(padding=self.padding + 12)

            bg_script_bgn: int = block_off + self.PFAT_INT_HDR_LEN
            bg_script_end: int = bg_script_bgn + bg_hdr.ScriptSize

            bg_data_bgn: int = bg_script_end
            bg_data_end: int = bg_data_bgn + bg_hdr.DataSize

            bg_data_bin: bytes = pfat_buffer[bg_data_bgn:bg_data_end]

            block_off = bg_data_end  # Assume next block starts at data end

            is_sfam, _, _, _, _ = bg_hdr.get_flags()  # SFAM, ProtectEC, GFXMitDis, FTU, Reserved

            if is_sfam:
                printer(message=f'Intel BIOS Guard {block_status} Signature:\n', padding=self.padding + 8)

                # Manual BIOS Guard Signature length detection from Header pattern (e.g. Panasonic)
                if bg_sign_len == 0:
                    bg_sign_sig: bytes = bg_hdr.get_hdr_marker()
                    bg_sign_lim: int = bg_data_end + self.PFAT_INT_SIG_MAX_LEN + len(bg_sign_sig)
                    bg_sign_len = pfat_buffer.find(bg_sign_sig, bg_data_end, bg_sign_lim) - bg_data_end

                # Adjust next block to start after current block Data + Signature
                block_off += self.parse_bg_sign(input_data=pfat_buffer, sign_offset=bg_data_end,
                                                sign_length=bg_sign_len, print_info=True, padding=self.padding + 12)

            printer(message=f'Intel BIOS Guard {block_status} Script:\n', padding=self.padding + 8)

            _ = self.parse_bg_script(script_data=pfat_buffer[bg_script_bgn:bg_script_end], padding=self.padding + 12)

            with open(file_path, 'ab') as out_dat:
                out_dat.write(bg_data_bin)

            all_blocks_dict[file_index] += bg_data_bin

            if block_index + 1 == block_count:
                ami_pfat_extract: AmiPfatExtract = AmiPfatExtract(
                    input_object=all_blocks_dict[file_index], extract_path=extract_folder(file_path),
                    padding=self.padding + 8)

                if ami_pfat_extract.check_format():
                    ami_pfat_extract.parse_format()

        pfat_oob_data: bytes = pfat_buffer[block_off:]  # Store out-of-bounds data after the end of PFAT files

        pfat_oob_name: str = self._get_file_name(index=file_count + 1, name=f'{extract_name}_OOB.bin')

        pfat_oob_path: str = os.path.join(self.extract_path, pfat_oob_name)

        with open(pfat_oob_path, 'wb') as out_oob:
            out_oob.write(pfat_oob_data)

        ami_pfat_extract = AmiPfatExtract(
            input_object=pfat_oob_data, extract_path=extract_folder(pfat_oob_path), padding=self.padding)

        if ami_pfat_extract.check_format():
            ami_pfat_extract.parse_format()

        in_all_data: bytes = b''.join([block[1] for block in sorted(all_blocks_dict.items())])

        in_all_name: str = self._get_file_name(index=0, name=f'{extract_name}_ALL.bin')

        in_all_path: str = os.path.join(self.extract_path, in_all_name)

        with open(in_all_path, 'wb') as out_all:
            out_all.write(in_all_data + pfat_oob_data)

        return True

    def parse_bg_sign(self, input_data: bytes, sign_offset: int, sign_length: int = 0,
                      print_info: bool = False, padding: int = 0) -> int:
        """ Process Intel BIOS Guard Signature """

        bg_sig_hdr: Any = ctypes_struct(buffer=input_data, start_offset=sign_offset,
                                        class_object=IntelBiosGuardSignatureHeader)

        if bg_sig_hdr.Unknown0 == 1:
            bg_sig_rsa_struct: Any = IntelBiosGuardSignatureRsa2k  # Unknown0/Unknown1 = 1,1
        elif bg_sig_hdr.Unknown0 in (2, 3):
            bg_sig_rsa_struct = IntelBiosGuardSignatureRsa3k  # Unknown0/Unknown1 = 2/3, 3/5, 3/6
        elif sign_length == self.PFAT_INT_SIG_HDR_LEN + self.PFAT_INT_SIG_R2K_LEN:
            bg_sig_rsa_struct = IntelBiosGuardSignatureRsa2k

            printer(message='Warning: Detected Intel BIOS Guard Signature 2K length via pattern!\n',
                    padding=padding, new_line=False)
        elif sign_length == self.PFAT_INT_SIG_HDR_LEN + self.PFAT_INT_SIG_R3K_LEN:
            bg_sig_rsa_struct = IntelBiosGuardSignatureRsa3k

            printer(message='Warning: Detected Intel BIOS Guard Signature 3K length via pattern!\n',
                    padding=padding, new_line=False)
        else:
            bg_sig_rsa_struct = IntelBiosGuardSignatureRsa3k

            printer(message='Error: Could not detect Intel BIOS Guard Signature length, assuming 3K!\n',
                    padding=padding, new_line=False)

        bg_sig_rsa: Any = ctypes_struct(buffer=input_data, start_offset=sign_offset + self.PFAT_INT_SIG_HDR_LEN,
                                        class_object=bg_sig_rsa_struct)

        if print_info:
            bg_sig_hdr.struct_print(padding=padding)

            bg_sig_rsa.struct_print(padding=padding)

        # Total size of Signature Header and RSA Structure
        return self.PFAT_INT_SIG_HDR_LEN + ctypes.sizeof(bg_sig_rsa_struct)

    @staticmethod
    def _get_ami_pfat(input_object: str | bytes | bytearray) -> bytes:
        """ Get actual AMI BIOS Guard buffer """

        input_buffer: bytes = file_to_bytes(in_object=input_object)

        match: re.Match[bytes] | None = PAT_AMI_PFAT.search(input_buffer)

        return input_buffer[match.start() - 0x8:] if match else b''

    @staticmethod
    def _get_file_name(index: int, name: str) -> str:
        """ Create AMI BIOS Guard output filename """

        return safe_name(in_name=f'{index:02d} -- {name}')

    @staticmethod
    def parse_bg_script(script_data: bytes, padding: int = 0) -> int:
        """ Process Intel BIOS Guard Script """

        is_opcode_div: bool = len(script_data) % 8 == 0

        if not is_opcode_div:
            printer(message='Error: BIOS Guard script is not divisible by OpCode length!',
                    padding=padding, new_line=False)

            return 1

        is_begin_end: bool = script_data[:8] + script_data[-8:] == b'\x01' + b'\x00' * 7 + b'\xFF' + b'\x00' * 7

        if not is_begin_end:
            printer(message='Error: BIOS Guard script lacks Begin and/or End OpCodes!',
                    padding=padding, new_line=False)

            return 2

        big_script: Type | None = big_script_tool()

        if not big_script:
            printer(message='Note: BIOS Guard Script Tool optional dependency is missing!',
                    padding=padding, new_line=False)

            return 3

        script: list[str] = big_script(code_bytes=script_data).to_string().replace('\t', '    ').split('\n')

        for opcode in script:
            if opcode.endswith(('begin', 'end')):
                spacing: int = padding
            elif opcode.endswith(':'):
                spacing = padding + 4
            else:
                spacing = padding + 12

            operands: list[str] = [operand for operand in opcode.split(' ') if operand]

            # Largest opcode length is 11 (erase64kblk) and largest operand length is 10 (0xAABBCCDD).
            printer(message=f'{operands[0]:11s}{"".join((f" {o:10s}" for o in operands[1:]))}',
                    padding=spacing, new_line=False)

        return 0

    def _parse_pfat_hdr(self, buffer: bytes | bytearray, padding: int = 0) -> tuple:
        """ Parse AMI BIOS Guard Header """

        block_all: list = []

        pfat_hdr: Any = ctypes_struct(buffer=buffer, start_offset=0, class_object=AmiBiosGuardHeader)

        hdr_size: int = pfat_hdr.Size

        hdr_data: bytes = buffer[self.PFAT_AMI_HDR_LEN:hdr_size]

        hdr_text: list[str] = hdr_data.decode('utf-8').splitlines()

        printer(message='AMI BIOS Guard Header:\n', padding=padding)

        pfat_hdr.struct_print(padding=padding + 4)

        hdr_title, *hdr_files = hdr_text

        files_count: int = len(hdr_files)

        hdr_tag, *hdr_indexes = hdr_title.split('II')

        printer(message=hdr_tag + '\n', padding=padding + 4)

        bgt_indexes: list = [int(h, 16) for h in re.findall(r'.{1,4}', hdr_indexes[0])] if hdr_indexes else []

        for index, entry in enumerate(hdr_files):
            entry_parts: list = entry.split(';')

            info: list = entry_parts[0].split()

            name: str = entry_parts[1]

            flags: int = int(info[0])

            param: str = info[1]

            count: int = int(info[2])

            order: str = to_ordinal(in_number=(bgt_indexes[index] if bgt_indexes else index) + 1)

            desc = f'{name} (Index: {index + 1:02d}, Flash: {order}, ' \
                   f'Parameter: {param}, Flags: 0x{flags:X}, Blocks: {count})'

            block_all += [(desc, name, order, param, flags, index, i, count) for i in range(count)]

        for block in block_all:
            if block[6] == 0:
                printer(message=block[0], padding=padding + 8, new_line=False)

        return block_all, hdr_size, files_count



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import re

from typing import Final

PAT_AMI_PFAT: Final[re.Pattern[bytes]] = re.compile(
    br'_AMIPFAT.AMI_BIOS_GUARD_FLASH_CONFIGURATIONS',
    flags=re.DOTALL
)

PAT_AMI_UCP: Final[re.Pattern[bytes]] = re.compile(
    br'@(UAF|HPU).{12}@',
    flags=re.DOTALL
)

PAT_APPLE_ROM_VER: Final[re.Pattern[bytes]] = re.compile(
    br'Apple ROM Version\x0A\x20{2}'
)

PAT_APPLE_IM4P: Final[re.Pattern[bytes]] = re.compile(
    br'\x16\x04IM4P\x16\x04mefi'
)

PAT_APPLE_PBZX: Final[re.Pattern[bytes]] = re.compile(
    br'pbzx'
)

PAT_AWARD_LZH: Final[re.Pattern[bytes]] = re.compile(
    br'-lh[04567]-'
)

PAT_DELL_FTR: Final[re.Pattern[bytes]] = re.compile(
    br'\xEE\xAA\xEE\x8F\x49\x1B\xE8\xAE\x14\x37\x90'
)

PAT_DELL_HDR: Final[re.Pattern[bytes]] = re.compile(
    br'\xEE\xAA\x76\x1B\xEC\xBB\x20\xF1\xE6\x51.\x78\x9C',
    flags=re.DOTALL
)

PAT_DELL_PKG: Final[re.Pattern[bytes]] = re.compile(
    br'\x72\x13\x55\x00.{45}7zXZ',
    flags=re.DOTALL
)

PAT_FUJITSU_SFX: Final[re.Pattern[bytes]] = re.compile(
    br'FjSfxBinay\xB2\xAC\xBC\xB9\xFF{4}.{4}\xFF{4}.{4}\xFF{4}\xFC\xFE',
    flags=re.DOTALL
)

PAT_INSYDE_IFL: Final[re.Pattern[bytes]] = re.compile(
    br'\$_IFLASH'
)

PAT_INSYDE_SFX: Final[re.Pattern[bytes]] = re.compile(
    br'\x0D\x0A;!@InstallEnd@!\x0D\x0A(7z\xBC\xAF\x27|\x6E\xF4\x79\x5F\x4E)'
)

PAT_INTEL_ENGINE: Final[re.Pattern[bytes]] = re.compile(
    br'\x04\x00{3}[\xA1\xE1]\x00{3}.{8}\x86\x80.{9}\x00\$((MN2)|(MAN))',
    flags=re.DOTALL
)

PAT_INTEL_FD: Final[re.Pattern[bytes]] = re.compile(
    br'\x5A\xA5\xF0\x0F.{172}\xFF{16}',
    flags=re.DOTALL
)

PAT_INTEL_IBIOSI: Final[re.Pattern[bytes]] = re.compile(
    br'\$IBIOSI\$.{16}\x2E\x00.{6}\x2E\x00.{8}\x2E\x00.{6}\x2E\x00.{20}\x00{2}',
    flags=re.DOTALL
)

PAT_MICROSOFT_CAB: Final[re.Pattern[bytes]] = re.compile(
    br'MSCF\x00{4}'
)

PAT_MICROSOFT_MZ: Final[re.Pattern[bytes]] = re.compile(
    br'MZ'
)

PAT_MICROSOFT_PE: Final[re.Pattern[bytes]] = re.compile(
    br'PE\x00{2}'
)

PAT_PHOENIX_TDK: Final[re.Pattern[bytes]] = re.compile(
    br'\$PACK\x00{3}..\x00{2}.\x00{3}',
    flags=re.DOTALL
)

PAT_PORTWELL_EFI: Final[re.Pattern[bytes]] = re.compile(
    br'<U{2}>'
)

PAT_TOSHIBA_COM: Final[re.Pattern[bytes]] = re.compile(
    br'\x00{2}[\x00-\x02]BIOS.{20}[\x00\x01]',
    flags=re.DOTALL
)

PAT_VAIO_CAB: Final[re.Pattern[bytes]] = re.compile(
    br'\xB2\xAC\xBC\xB9\xFF{4}.{4}\xFF{4}.{4}\xFF{4}\xFC\xFE',
    flags=re.DOTALL
)

PAT_VAIO_CFG: Final[re.Pattern[bytes]] = re.compile(
    br'\[Setting]\x0D\x0A'
)

PAT_VAIO_CHK: Final[re.Pattern[bytes]] = re.compile(
    br'\x0AUseVAIOCheck='
)

PAT_VAIO_EXT: Final[re.Pattern[bytes]] = re.compile(
    br'\x0AExtractPathByUser='
)



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Fujitsu SFX Extractor
Fujitsu SFX BIOS Extractor
Copyright (C) 2019-2024 Plato Mavropoulos
"""

import os
import re

from biosutilities.common.compression import is_szip_supported, szip_decompress
from biosutilities.common.paths import delete_file, make_dirs
from biosutilities.common.patterns import PAT_FUJITSU_SFX
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility


class FujitsuSfxExtract(BIOSUtility):
    """ Fujitsu SFX BIOS Extractor """

    TITLE: str = 'Fujitsu SFX BIOS Extractor'

    def check_format(self) -> bool:
        """ Check if input is Fujitsu SFX image """

        return bool(PAT_FUJITSU_SFX.search(self.input_buffer))

    def parse_format(self) -> bool:
        """ Parse & Extract Fujitsu SFX image """

        # Microsoft CAB Header XOR 0xFF
        match_cab: re.Match[bytes] | None = PAT_FUJITSU_SFX.search(self.input_buffer)

        if not match_cab:
            return False

        printer(message='Detected obfuscated CAB archive!', padding=self.padding)

        # Microsoft CAB Header XOR 0xFF starts after "FjSfxBinay" signature
        cab_start: int = match_cab.start() + 0xA

        # Get LE XOR-ed CAB size
        cab_size: int = int.from_bytes(self.input_buffer[cab_start + 0x8:cab_start + 0xC], byteorder='little')

        # Create CAB size XOR value
        xor_size: int = int.from_bytes(b'\xFF' * 0x4, byteorder='little')

        # Perform XOR 0xFF and get actual CAB size
        cab_size ^= xor_size

        printer(message='Removing obfuscation...', padding=self.padding + 4)

        # Get BE XOR-ed CAB data
        cab_data: int = int.from_bytes(self.input_buffer[cab_start:cab_start + cab_size], byteorder='big')

        # Create CAB data XOR value
        xor_data: int = int.from_bytes(b'\xFF' * cab_size, byteorder='big')

        # Perform XOR 0xFF and get actual CAB data
        raw_data: bytes = (cab_data ^ xor_data).to_bytes(cab_size, 'big')

        printer(message='Extracting archive...', padding=self.padding + 4)

        make_dirs(in_path=self.extract_path)

        cab_path: str = os.path.join(self.extract_path, 'FjSfxBinay.cab')

        # Create temporary CAB archive
        with open(cab_path, 'wb') as cab_file_object:
            cab_file_object.write(raw_data)

        if is_szip_supported(in_path=cab_path):
            if szip_decompress(in_path=cab_path, out_path=self.extract_path, in_name='FjSfxBinay CAB',
                               padding=self.padding + 8, check=True):
                delete_file(in_path=cab_path)
            else:
                return False
        else:
            return False

        return True



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import sys
import platform

from biosutilities.common.texts import to_string


def system_platform() -> tuple[str, bool, bool]:
    """ Get OS platform """

    sys_os: str = platform.system()

    is_win: bool = sys_os == 'Windows'

    is_lnx: bool = sys_os in ('Linux', 'Darwin')

    return sys_os, is_win, is_lnx


def python_version() -> tuple:
    """ Get Python version """

    return sys.version_info


def printer(message: str | list | tuple | None = None, padding: int = 0, new_line: bool = True,
            sep_char: str = ' ', strip: bool = False) -> None:
    """ Show message(s), controlling padding, newline, stripping, pausing & separating """

    message_string: str = to_string(in_object='' if message is None else message, sep_char=sep_char)

    message_output: str = '\n' if new_line else ''

    for message_line_index, message_line_text in enumerate(message_string.split('\n')):
        line_new: str = '' if message_line_index == 0 else '\n'

        line_text: str = message_line_text.strip() if strip else message_line_text

        message_output += f'{line_new}{" " * padding}{line_text}'

    print(message_output)



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import subprocess

from typing import Final

from biosutilities.common.externals import szip_path, tiano_path
from biosutilities.common.paths import is_dir, is_file_read, is_empty_dir, path_size
from biosutilities.common.system import printer
from biosutilities.common.texts import file_to_bytes

# 7-Zip switches to auto rename, ignore passwords, ignore prompts, ignore wildcards,
# eliminate root duplication, set UTF-8 charset, suppress stdout, suppress stderr,
# suppress progress, disable headers, disable progress, disable output logging
SZIP_COMMON: Final[list[str]] = ['-aou', '-p', '-y', '-spd', '-spe', '-sccUTF-8',
                                 '-bso0', '-bse0', '-bsp0', '-ba', '-bd', '-bb0']

# Success exit codes (0 = OK, 1 = Warnings)
SZIP_SUCCESS: Final[list[int]] = [0, 1]


def szip_switches(in_switches: list[str]) -> list[str]:
    """ Generate 7-Zip command line switches """

    common_switches: list[str] = SZIP_COMMON

    for in_switch in in_switches:
        for sw_pattern in ('-p', '-ao', '-bs', '-bb', '-scc'):
            if in_switch.startswith(sw_pattern):
                common_switches = [sw for sw in common_switches if not sw.startswith(sw_pattern)]

                break

    return [*set(common_switches + in_switches), '--']


def is_szip_successful(exit_code: int) -> bool:
    """ Check 7-Zip success exit codes """

    if exit_code in SZIP_SUCCESS:
        return True

    return False


def is_szip_supported(in_path: str, args: list | None = None) -> bool:
    """ Check if file is 7-Zip supported """

    szip_a: list[str] = [] if args is None else args

    szip_c: list[str] = [szip_path(), 't', *szip_switches(in_switches=[*szip_a]), in_path]

    szip_t: subprocess.CompletedProcess[bytes] = subprocess.run(szip_c, check=False, stdout=subprocess.DEVNULL)

    return is_szip_successful(exit_code=szip_t.returncode)


def szip_decompress(in_path: str, out_path: str, in_name: str = 'archive', padding: int = 0, args: list | None = None,
                    check: bool = False, silent: bool = False) -> bool:
    """ Archive decompression via 7-Zip """

    szip_a: list[str] = [] if args is None else args

    szip_c: list[str] = [szip_path(), 'x', *szip_switches(in_switches=[*szip_a, f'-o{out_path}']), in_path]

    szip_x: subprocess.CompletedProcess[bytes] = subprocess.run(szip_c, check=False, stdout=subprocess.DEVNULL)

    szip_s: bool = is_szip_successful(exit_code=szip_x.returncode) if check else True

    if szip_s and is_dir(in_path=out_path) and not is_empty_dir(in_path=out_path):
        if not silent:
            printer(message=f'Successful {in_name} decompression via 7-Zip!', padding=padding)

        return True

    return False


def efi_header_info(in_object: str | bytes | bytearray) -> dict[str, int]:
    """ Get EFI compression sizes from header """

    efi_data: bytes = file_to_bytes(in_object=in_object)

    size_compressed: int = int.from_bytes(efi_data[0x0:0x4], byteorder='little')

    size_decompressed: int = int.from_bytes(efi_data[0x4:0x8], byteorder='little')

    return {'size_compressed': size_compressed, 'size_decompressed': size_decompressed}


def is_efi_compressed(in_object: str | bytes | bytearray, strict: bool = True) -> bool:
    """ Check if data is EFI compressed, controlling EOF padding """

    efi_data: bytes = file_to_bytes(in_object=in_object)

    efi_sizes: dict[str, int] = efi_header_info(in_object=efi_data)

    check_diff: bool = efi_sizes['size_compressed'] < efi_sizes['size_decompressed']

    if strict:
        check_size: bool = efi_sizes['size_compressed'] + 0x8 == len(efi_data)
    else:
        check_size = efi_sizes['size_compressed'] + 0x8 <= len(efi_data)

    return check_diff and check_size


def efi_decompress(in_path: str, out_path: str, padding: int = 0, silent: bool = False,
                   comp_type: str = '--uefi') -> bool:
    """ EFI/Tiano Decompression via TianoCompress """

    tiano_c: list[str] = [tiano_path(), '-d', in_path, '-o', out_path, '-q', comp_type]

    tiano_x: subprocess.CompletedProcess[bytes] = subprocess.run(tiano_c, check=False, stdout=subprocess.DEVNULL)

    if tiano_x.returncode == 0 and is_file_read(in_path=out_path):
        if efi_header_info(in_object=in_path)['size_decompressed'] == path_size(in_path=out_path):
            if not silent:
                printer(message='Successful EFI decompression via TianoCompress!', padding=padding)

            return True

    return False



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Portwell EFI Extract
Portwell EFI Update Extractor
Copyright (C) 2021-2024 Plato Mavropoulos
"""

import logging
import os

from re import Match
from typing import Final

from pefile import PE

from biosutilities.common.compression import efi_decompress, is_efi_compressed
from biosutilities.common.paths import delete_file, make_dirs, rename_file, safe_name
from biosutilities.common.executables import ms_pe
from biosutilities.common.patterns import PAT_MICROSOFT_MZ, PAT_PORTWELL_EFI
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility


class PortwellEfiExtract(BIOSUtility):
    """ Portwell EFI Update Extractor """

    TITLE: str = 'Portwell EFI Update Extractor'

    FILE_NAMES: Final[dict[int, str]] = {
        0: 'Flash.efi',
        1: 'Fparts.txt',
        2: 'Update.nsh',
        3: 'Temp.bin',
        4: 'SaveDmiData.efi'
    }

    def check_format(self) -> bool:
        """ Check if input is Portwell EFI executable """

        try:
            pe_buffer: bytes = self._get_portwell_pe(in_buffer=self.input_buffer)[1]
        except Exception as error:  # pylint: disable=broad-except
            logging.debug('Error: Could not check if input is Portwell EFI executable: %s', error)

            return False

        # EFI images start with PE Header MZ
        if PAT_MICROSOFT_MZ.search(self.input_buffer[:0x2]):
            # Portwell EFI files start with <UU>
            if PAT_PORTWELL_EFI.search(pe_buffer[:0x4]):
                return True

        return False

    def parse_format(self) -> bool:
        """ Parse & Extract Portwell UEFI Unpacker """

        # Initialize EFI Payload file chunks
        efi_files: list[bytes] = []

        make_dirs(in_path=self.extract_path)

        pe_file, pe_data = self._get_portwell_pe(in_buffer=self.input_buffer)

        efi_title: str = self._get_unpacker_tag(input_buffer=self.input_buffer, pe_file=pe_file)

        printer(message=efi_title, padding=self.padding)

        # Split EFI Payload into <UU> file chunks
        efi_list: list[Match[bytes]] = list(PAT_PORTWELL_EFI.finditer(pe_data))

        for idx, val in enumerate(efi_list):
            efi_bgn: int = val.end()
            efi_end: int = len(pe_data) if idx == len(efi_list) - 1 else efi_list[idx + 1].start()

            efi_files.append(pe_data[efi_bgn:efi_end])

        self._parse_efi_files(extract_path=self.extract_path, efi_files=efi_files, padding=self.padding)

        return True

    @staticmethod
    def _get_portwell_pe(in_buffer: bytes) -> tuple:
        """ Get PE of Portwell EFI executable """

        # Analyze EFI Portable Executable (PE)
        pe_file: PE | None = ms_pe(in_buffer, silent=True)

        # Skip EFI executable
        # pylint: disable=no-member
        pe_data: bytes = in_buffer[pe_file.OPTIONAL_HEADER.SizeOfImage:]  # type: ignore

        return pe_file, pe_data

    @staticmethod
    def _get_unpacker_tag(input_buffer: bytes | bytearray, pe_file: PE) -> str:
        """ Get Portwell UEFI Unpacker tag """

        unpacker_tag_txt: str = 'UEFI Unpacker'

        for pe_section in pe_file.sections:
            # Unpacker Tag, Version, Strings etc. are found in .data PE section
            if pe_section.Name.startswith(b'.data'):
                pe_data_bgn: int = pe_section.PointerToRawData
                pe_data_end: int = pe_data_bgn + pe_section.SizeOfRawData

                # Decode any valid UTF-16 .data PE section info to a parsable text buffer
                pe_data_txt: str = input_buffer[pe_data_bgn:pe_data_end].decode('utf-16', 'ignore')

                # Search .data for UEFI Unpacker tag
                unpacker_tag_bgn: int = pe_data_txt.find(unpacker_tag_txt)

                if unpacker_tag_bgn != -1:
                    unpacker_tag_len: int = pe_data_txt[unpacker_tag_bgn:].find('=')

                    if unpacker_tag_len != -1:
                        unpacker_tag_end: int = unpacker_tag_bgn + unpacker_tag_len
                        unpacker_tag_raw: str = pe_data_txt[unpacker_tag_bgn:unpacker_tag_end]

                        # Found full UEFI Unpacker tag, store and slightly beautify the resulting text
                        unpacker_tag_txt = unpacker_tag_raw.strip().replace('   ', ' ').replace('<', ' <')

                break  # Found PE .data section, skip the rest

        return unpacker_tag_txt

    def _parse_efi_files(self, extract_path: str, efi_files: list[bytes], padding: int) -> None:
        """ Process Portwell UEFI Unpacker payload files """

        for file_index, file_data in enumerate(efi_files):
            if file_data in (b'', b'NULL'):
                continue  # Skip empty/unused files

            # Assign Name to EFI file
            file_name: str = self.FILE_NAMES.get(file_index, f'Unknown_{file_index}.bin')

            # Print EFI file name, indicate progress
            printer(message=f'[{file_index}] {file_name}', padding=padding + 4)

            if file_name.startswith('Unknown_'):
                printer(message=f'Note: Detected new Portwell EFI file ID {file_index}!', padding=padding + 8)

            # Store EFI file output path
            file_path: str = os.path.join(extract_path, safe_name(in_name=file_name))

            # Store EFI file data to drive
            with open(file_path, 'wb') as out_file:
                out_file.write(file_data)

            # Attempt to detect EFI compression & decompress when applicable
            if is_efi_compressed(in_object=file_data):
                # Store temporary compressed file name
                file_path_temp: str = f'{file_path}.temp'

                # Rename initial/compressed file
                rename_file(in_path=file_path, in_dest=file_path_temp)

                # Successful decompression, delete compressed file
                if efi_decompress(in_path=file_path_temp, out_path=file_path, padding=padding + 8):
                    delete_file(in_path=file_path_temp)



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Apple PBZX Extract
Apple EFI PBZX Extractor
Copyright (C) 2021-2024 Plato Mavropoulos
"""

import ctypes
import logging
import lzma
import os

from typing import Any, Final

from biosutilities.common.compression import is_szip_supported, szip_decompress
from biosutilities.common.paths import delete_file, is_file_read, make_dirs, path_stem
from biosutilities.common.patterns import PAT_APPLE_PBZX
from biosutilities.common.structs import ctypes_struct, UINT32
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility


class PbzxChunk(ctypes.BigEndianStructure):
    """ PBZX Chunk Header """

    _pack_ = 1
    _fields_ = [
        ('Reserved0',       UINT32),        # 0x00
        ('InitSize',        UINT32),        # 0x04
        ('Reserved1',       UINT32),        # 0x08
        ('CompSize',        UINT32)         # 0x0C
        # 0x10
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Reserved 0     :', f'0x{self.Reserved0:X}'], padding=padding, new_line=False)
        printer(message=['Initial Size   :', f'0x{self.InitSize:X}'], padding=padding, new_line=False)
        printer(message=['Reserved 1     :', f'0x{self.Reserved1:X}'], padding=padding, new_line=False)
        printer(message=['Compressed Size:', f'0x{self.CompSize:X}'], padding=padding, new_line=False)


class AppleEfiPbzxExtract(BIOSUtility):
    """ Apple EFI PBZX Extractor """

    TITLE: str = 'Apple EFI PBZX Extractor'

    PBZX_CHUNK_HDR_LEN: Final[int] = ctypes.sizeof(PbzxChunk)

    def check_format(self) -> bool:
        """ Check if input is Apple PBZX image """

        if isinstance(self.input_object, str) and is_file_read(in_path=self.input_object):
            with open(self.input_object, 'rb') as input_object:
                check_buffer: bytes = input_object.read(4)
        else:
            check_buffer = self.input_buffer[:4]

        return bool(PAT_APPLE_PBZX.search(check_buffer, 0, 4))

    def parse_format(self) -> bool:
        """ Parse & Extract Apple PBZX image """

        make_dirs(in_path=self.extract_path)

        cpio_bin: bytes = b''  # Initialize PBZX > CPIO Buffer

        cpio_len: int = 0x0  # Initialize PBZX > CPIO Length

        chunk_off: int = 0xC  # First PBZX Chunk starts at 0xC

        while chunk_off < len(self.input_buffer):
            chunk_hdr: Any = ctypes_struct(buffer=self.input_buffer, start_offset=chunk_off, class_object=PbzxChunk)

            printer(message=f'PBZX Chunk at 0x{chunk_off:08X}\n', padding=self.padding)

            chunk_hdr.struct_print(padding=self.padding + 4)

            # PBZX Chunk data starts after its Header
            comp_bgn: int = chunk_off + self.PBZX_CHUNK_HDR_LEN

            # To avoid a potential infinite loop, double-check Compressed Size
            comp_end: int = comp_bgn + max(chunk_hdr.CompSize, self.PBZX_CHUNK_HDR_LEN)

            comp_bin: bytes = self.input_buffer[comp_bgn:comp_end]

            try:
                # Attempt XZ decompression, if applicable to Chunk data
                cpio_bin += lzma.LZMADecompressor().decompress(comp_bin)

                printer(message='Successful LZMA decompression!', padding=self.padding + 8)
            except Exception as error:  # pylint: disable=broad-except
                logging.debug('Error: Failed to LZMA decompress PBZX Chunk 0x%X: %s', chunk_off, error)

                # Otherwise, Chunk data is not compressed
                cpio_bin += comp_bin

            # Final CPIO size should match the sum of all Chunks > Initial Size
            cpio_len += chunk_hdr.InitSize

            # Next Chunk starts at the end of current Chunk's data
            chunk_off = comp_end

        # Check that CPIO size is valid based on all Chunks > Initial Size
        if cpio_len != len(cpio_bin):
            printer(message='Error: Unexpected CPIO archive size!', padding=self.padding)

            return False

        cpio_name: str = path_stem(in_path=self.input_object) if isinstance(self.input_object, str) else 'Payload'

        cpio_path: str = os.path.join(self.extract_path, f'{cpio_name}.cpio')

        with open(cpio_path, 'wb') as cpio_object:
            cpio_object.write(cpio_bin)

        # Decompress PBZX > CPIO archive with 7-Zip
        if is_szip_supported(in_path=cpio_path, args=['-tCPIO']):
            if szip_decompress(in_path=cpio_path, out_path=self.extract_path, in_name='CPIO',
                               padding=self.padding, args=['-tCPIO']):
                delete_file(in_path=cpio_path)  # Successful extraction, delete PBZX > CPIO archive
            else:
                return False
        else:
            return False

        return True



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Dell PFS Extract
Dell PFS Update Extractor
Copyright (C) 2018-2024 Plato Mavropoulos
"""

import contextlib
import ctypes
import io
import lzma
import os
import zlib

from re import Match
from typing import Any, Final

from biosutilities.common.checksums import checksum_8_xor
from biosutilities.common.compression import is_szip_supported, szip_decompress
from biosutilities.common.paths import (delete_dirs, delete_file, path_files, is_file_read, make_dirs,
                                        path_name, path_parent, path_stem, safe_name)
from biosutilities.common.patterns import PAT_DELL_FTR, PAT_DELL_HDR, PAT_DELL_PKG
from biosutilities.common.structs import CHAR, ctypes_struct, UINT8, UINT16, UINT32, UINT64
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility
from biosutilities.common.texts import file_to_bytes, to_ordinal

from biosutilities.ami_pfat_extract import AmiPfatExtract, IntelBiosGuardHeader


class DellPfsHeader(ctypes.LittleEndianStructure):
    """ Dell PFS Header Structure """

    _pack_ = 1
    _fields_ = [
        ('Tag',             CHAR * 8),      # 0x00
        ('HeaderVersion',   UINT32),        # 0x08
        ('PayloadSize',     UINT32)         # 0x0C
        # 0x10
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Header Tag    :', self.Tag.decode('utf-8')], padding=padding, new_line=False)
        printer(message=['Header Version:', self.HeaderVersion], padding=padding, new_line=False)
        printer(message=['Payload Size  :', f'0x{self.PayloadSize:X}'], padding=padding, new_line=False)


class DellPfsFooter(ctypes.LittleEndianStructure):
    """ Dell PFS Footer Structure """

    _pack_ = 1
    _fields_ = [
        ('PayloadSize',     UINT32),        # 0x00
        ('Checksum',        UINT32),        # 0x04 ~CRC32 w/ Vector 0
        ('Tag',             CHAR * 8)       # 0x08
        # 0x10
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Payload Size    :', f'0x{self.PayloadSize:X}'], padding=padding, new_line=False)
        printer(message=['Payload Checksum:', f'0x{self.Checksum:08X}'], padding=padding, new_line=False)
        printer(message=['Footer Tag      :', self.Tag.decode('utf-8')], padding=padding, new_line=False)


class DellPfsEntryBase(ctypes.LittleEndianStructure):
    """ Dell PFS Entry Base Structure """

    _pack_ = 1
    _fields_ = [
        ('GUID',            UINT32 * 4),    # 0x00 Little Endian
        ('HeaderVersion',   UINT32),        # 0x10 1 or 2
        ('VersionType',     UINT8 * 4),     # 0x14
        ('Version',         UINT16 * 4),    # 0x18
        ('Reserved',        UINT64),        # 0x20
        ('DataSize',        UINT32),        # 0x28
        ('DataSigSize',     UINT32),        # 0x2C
        ('DataMetSize',     UINT32),        # 0x30
        ('DataMetSigSize',  UINT32)         # 0x34
        # 0x38 (parent class, base)
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        guid: str = f'{int.from_bytes(self.GUID, byteorder="little"):0{0x10 * 2}X}'
        unknown: str = f'{int.from_bytes(self.Unknown, byteorder="little"):0{len(self.Unknown) * 8}X}'
        version: str = DellPfsExtract.get_entry_ver(version_fields=self.Version, version_types=self.VersionType)

        printer(message=['Entry GUID             :', guid], padding=padding, new_line=False)
        printer(message=['Entry Version          :', self.HeaderVersion], padding=padding, new_line=False)
        printer(message=['Payload Version        :', version], padding=padding, new_line=False)
        printer(message=['Reserved               :', f'0x{self.Reserved:X}'], padding=padding, new_line=False)
        printer(message=['Payload Data Size      :', f'0x{self.DataSize:X}'], padding=padding, new_line=False)
        printer(message=['Payload Signature Size :', f'0x{self.DataSigSize:X}'], padding=padding, new_line=False)
        printer(message=['Metadata Data Size     :', f'0x{self.DataMetSize:X}'], padding=padding, new_line=False)
        printer(message=['Metadata Signature Size:', f'0x{self.DataMetSigSize:X}'], padding=padding, new_line=False)
        printer(message=['Unknown                :', f'0x{unknown}'], padding=padding, new_line=False)


class DellPfsEntryR1(DellPfsEntryBase):
    """ Dell PFS Entry Revision 1 Structure """

    _pack_ = 1
    _fields_ = [
        ('Unknown',         UINT32 * 4)     # 0x38
        # 0x48 (child class, R1)
    ]


class DellPfsEntryR2(DellPfsEntryBase):
    """ Dell PFS Entry Revision 2 Structure """

    _pack_ = 1
    _fields_ = [
        ('Unknown',         UINT32 * 8)     # 0x38
        # 0x58 (child class, R2)
    ]


class DellPfsInfo(ctypes.LittleEndianStructure):
    """ Dell PFS Information Header Structure """

    _pack_ = 1
    _fields_ = [
        ('HeaderVersion',   UINT32),        # 0x00
        ('GUID',            UINT32 * 4)     # 0x04 Little Endian
        # 0x14
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        guid: str = f'{int.from_bytes(self.GUID, byteorder="little"):0{0x10 * 2}X}'

        printer(message=['Info Version:', self.HeaderVersion], padding=padding, new_line=False)
        printer(message=['Entry GUID  :', guid], padding=padding, new_line=False)


class DellPfsName(ctypes.LittleEndianStructure):
    """ Dell PFS FileName Header Structure """

    _pack_ = 1
    _fields_ = [
        ('Version',         UINT16 * 4),    # 0x00
        ('VersionType',     UINT8 * 4),     # 0x08
        ('CharacterCount',  UINT16)         # 0x0C UTF-16 2-byte Characters
        # 0x0E
    ]

    def struct_print(self, name: str, padding: int = 0) -> None:
        """ Display structure information """

        version: str = DellPfsExtract.get_entry_ver(self.Version, self.VersionType)

        printer(message=['Payload Version:', version], padding=padding, new_line=False)
        printer(message=['Character Count:', self.CharacterCount], padding=padding, new_line=False)
        printer(message=['Payload Name   :', name], padding=padding, new_line=False)


class DellPfsMetadata(ctypes.LittleEndianStructure):
    """ Dell PFS Metadata Header Structure """

    _pack_ = 1
    _fields_ = [
        ('ModelIDs',        CHAR * 501),    # 0x000
        ('FileName',        CHAR * 100),    # 0x1F5
        ('FileVersion',     CHAR * 33),     # 0x259
        ('Date',            CHAR * 33),     # 0x27A
        ('Brand',           CHAR * 80),     # 0x29B
        ('ModelFile',       CHAR * 80),     # 0x2EB
        ('ModelName',       CHAR * 100),    # 0x33B
        ('ModelVersion',    CHAR * 33)      # 0x39F
        # 0x3C0
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        model_ids: str = self.ModelIDs.decode('utf-8').removesuffix(',END')

        printer(message=['Model IDs    :', model_ids], padding=padding, new_line=False)
        printer(message=['File Name    :', self.FileName.decode('utf-8')], padding=padding, new_line=False)
        printer(message=['File Version :', self.FileVersion.decode('utf-8')], padding=padding, new_line=False)
        printer(message=['Date         :', self.Date.decode('utf-8')], padding=padding, new_line=False)
        printer(message=['Brand        :', self.Brand.decode('utf-8')], padding=padding, new_line=False)
        printer(message=['Model File   :', self.ModelFile.decode('utf-8')], padding=padding, new_line=False)
        printer(message=['Model Name   :', self.ModelName.decode('utf-8')], padding=padding, new_line=False)
        printer(message=['Model Version:', self.ModelVersion.decode('utf-8')], padding=padding, new_line=False)


class DellPfsPfatMetadata(ctypes.LittleEndianStructure):
    """ Dell PFS BIOS Guard Metadata Structure """

    _pack_ = 1
    _fields_ = [
        ('Address',         UINT32),        # 0x00
        ('Unknown0',        UINT32),        # 0x04
        ('Offset',          UINT32),        # 0x08 Matches BG Script > I0
        ('DataSize',        UINT32),        # 0x0C Matches BG Script > I2 & Header > Data Size
        ('Unknown1',        UINT32),        # 0x10
        ('Unknown2',        UINT32),        # 0x14
        ('Unknown3',        UINT8)          # 0x18
        # 0x19
    ]

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Address  :', f'0x{self.Address:X}'], padding=padding, new_line=False)
        printer(message=['Unknown 0:', f'0x{self.Unknown0:X}'], padding=padding, new_line=False)
        printer(message=['Offset   :', f'0x{self.Offset:X}'], padding=padding, new_line=False)
        printer(message=['Length   :', f'0x{self.DataSize:X}'], padding=padding, new_line=False)
        printer(message=['Unknown 1:', f'0x{self.Unknown1:X}'], padding=padding, new_line=False)
        printer(message=['Unknown 2:', f'0x{self.Unknown2:X}'], padding=padding, new_line=False)
        printer(message=['Unknown 3:', f'0x{self.Unknown3:X}'], padding=padding, new_line=False)


class DellPfsExtract(BIOSUtility):
    """ Dell PFS Update Extractor """

    TITLE: str = 'Dell PFS Update Extractor'

    PFS_HEAD_LEN: Final[int] = ctypes.sizeof(DellPfsHeader)
    PFS_FOOT_LEN: Final[int] = ctypes.sizeof(DellPfsFooter)
    PFS_INFO_LEN: Final[int] = ctypes.sizeof(DellPfsInfo)
    PFS_NAME_LEN: Final[int] = ctypes.sizeof(DellPfsName)
    PFS_META_LEN: Final[int] = ctypes.sizeof(DellPfsMetadata)
    PFS_PFAT_LEN: Final[int] = ctypes.sizeof(DellPfsPfatMetadata)
    PFAT_HDR_LEN: Final[int] = ctypes.sizeof(IntelBiosGuardHeader)

    def __init__(self, input_object: str | bytes | bytearray = b'', extract_path: str = '', padding: int = 0,
                 advanced: bool = False, structure: bool = False) -> None:
        super().__init__(input_object=input_object, extract_path=extract_path, padding=padding)

        self.advanced: bool = advanced
        self.structure: bool = structure

    def check_format(self) -> bool:
        """ Check if input is Dell PFS/PKG image """

        if self._is_pfs_pkg(input_object=self.input_buffer):
            return True

        if self._is_pfs_hdr(input_object=self.input_buffer) and self._is_pfs_ftr(input_object=self.input_buffer):
            return True

        return False

    def parse_format(self) -> bool:
        """ Parse & Extract Dell PFS Update image """

        make_dirs(in_path=self.extract_path)

        is_dell_pkg: bool = self._is_pfs_pkg(input_object=self.input_buffer)

        if is_dell_pkg:
            pfs_results: dict[str, bytes] = self._thinos_pkg_extract(
                input_object=self.input_buffer, extract_path=self.extract_path)
        else:
            pfs_results = {path_stem(in_path=self.input_object) if isinstance(self.input_object, str) and is_file_read(
                in_path=self.input_object) else 'Image': self.input_buffer}

        # Parse each Dell PFS image contained in the input file
        for pfs_index, (pfs_name, pfs_buffer) in enumerate(pfs_results.items(), start=1):
            # At ThinOS PKG packages, multiple PFS images may be included in separate model-named folders
            pfs_path: str = os.path.join(
                self.extract_path, f'{pfs_index} {pfs_name}') if is_dell_pkg else self.extract_path

            # Parse each PFS ZLIB section
            for zlib_offset in self._get_section_offsets(buffer=pfs_buffer):
                # Call the PFS ZLIB section parser function
                self._pfs_section_parse(zlib_data=pfs_buffer, zlib_start=zlib_offset, extract_path=pfs_path,
                                        pfs_name=pfs_name, pfs_index=pfs_index, pfs_count=1, is_rec=False,
                                        padding=self.padding)

        return True

    @staticmethod
    def _is_pfs_pkg(input_object: str | bytes | bytearray) -> bool:
        """
        The Dell ThinOS PKG update images usually contain multiple sections.
        Each section starts with a 0x30 header, which begins with pattern 72135500.
        The section length is found at 0x10-0x14 and its (optional) MD5 hash at 0x20-0x30.
        Section data can be raw or LZMA2 (7zXZ) compressed. The latter contains the PFS update image.
        """

        input_buffer: bytes = file_to_bytes(in_object=input_object)

        return bool(PAT_DELL_PKG.search(input_buffer))

    @staticmethod
    def _is_pfs_hdr(input_object: str | bytes | bytearray) -> bool:
        """
        The Dell PFS update images usually contain multiple sections.
        Each section is zlib-compressed with header pattern ********++EEAA761BECBB20F1E651--789C,
        where ******** is the zlib stream size, ++ is the section type and -- the header Checksum XOR 8.
        The "Firmware" section has type AA and its files are stored in PFS format.
        The "Utility" section has type BB and its files are stored in PFS, BIN or 7z formats.
        """

        input_buffer: bytes = file_to_bytes(in_object=input_object)

        return bool(PAT_DELL_HDR.search(input_buffer))

    @staticmethod
    def _is_pfs_ftr(input_object: str | bytes | bytearray) -> bool:
        """
        Each section is followed by the footer pattern ********EEAAEE8F491BE8AE143790--,
        where ******** is the zlib stream size and ++ the footer Checksum XOR 8.
        """

        input_buffer: bytes = file_to_bytes(in_object=input_object)

        return bool(PAT_DELL_FTR.search(input_buffer))

    def _thinos_pkg_extract(self, input_object: str | bytes | bytearray, extract_path: str) -> dict[str, bytes]:
        """ Extract Dell ThinOS PKG 7zXZ """

        input_buffer: bytes = file_to_bytes(in_object=input_object)

        # Initialize PFS results (Name: Buffer)
        pfs_results: dict[str, bytes] = {}

        # Search input image for ThinOS PKG 7zXZ header
        thinos_pkg_match: Match[bytes] | None = PAT_DELL_PKG.search(input_buffer)

        if not thinos_pkg_match:
            return pfs_results

        lzma_len_off: int = thinos_pkg_match.start() + 0x10
        lzma_len_int: int = int.from_bytes(input_buffer[lzma_len_off:lzma_len_off + 0x4], byteorder='little')
        lzma_bin_off: int = thinos_pkg_match.end() - 0x5

        lzma_bin_dat: bytes = input_buffer[lzma_bin_off:lzma_bin_off + lzma_len_int]

        # Check if the compressed 7zXZ stream is complete
        if len(lzma_bin_dat) != lzma_len_int:
            return pfs_results

        working_path: str = os.path.join(extract_path, 'THINOS_PKG_TEMP')

        make_dirs(in_path=working_path, delete=True)

        pkg_tar_path: str = os.path.join(working_path, 'THINOS_PKG.TAR')

        with open(pkg_tar_path, 'wb') as pkg_payload:
            pkg_payload.write(lzma.decompress(lzma_bin_dat))

        if is_szip_supported(in_path=pkg_tar_path, args=['-tTAR']):
            if szip_decompress(in_path=pkg_tar_path, out_path=working_path, in_name='TAR', padding=0,
                               args=['-tTAR'], check=True, silent=True):
                delete_file(in_path=pkg_tar_path)
            else:
                return pfs_results
        else:
            return pfs_results

        for pkg_file in path_files(in_path=working_path):
            if is_file_read(in_path=pkg_file):
                if self._is_pfs_hdr(input_object=pkg_file):
                    pfs_name: str = path_name(in_path=str(path_parent(in_path=pkg_file)))

                    pfs_results.update({pfs_name: file_to_bytes(in_object=pkg_file)})

        delete_dirs(in_path=working_path)

        return pfs_results

    @staticmethod
    def _get_section_offsets(buffer: bytes | bytearray) -> list[int]:
        """ Get PFS ZLIB Section Offsets """

        pfs_zlib_list: list[int] = []  # Initialize PFS ZLIB offset list

        pfs_zlib_init: list[Match[bytes]] = list(PAT_DELL_HDR.finditer(buffer))

        if not pfs_zlib_init:
            return pfs_zlib_list  # No PFS ZLIB detected

        # Remove duplicate/nested PFS ZLIB offsets
        for zlib_c in pfs_zlib_init:
            is_duplicate: bool = False  # Initialize duplicate/nested PFS ZLIB offset

            for zlib_o in pfs_zlib_init:
                zlib_o_size: int = int.from_bytes(buffer[zlib_o.start() - 0x5:zlib_o.start() - 0x1],
                                                  byteorder='little')

                # If current PFS ZLIB offset is within another PFS ZLIB range (start-end), set as duplicate
                if zlib_o.start() < zlib_c.start() < zlib_o.start() + zlib_o_size:
                    is_duplicate = True

            if not is_duplicate:
                pfs_zlib_list.append(zlib_c.start())

        return pfs_zlib_list

    def _pfs_section_parse(self, zlib_data: bytes | bytearray, zlib_start: int, extract_path: str, pfs_name: str,
                           pfs_index: int, pfs_count: int, is_rec: bool, padding: int = 0) -> None:
        """ Dell PFS ZLIB Section Parser """

        # Initialize PFS ZLIB-related error state
        is_zlib_error: bool = False

        # Byte before PFS ZLIB Section pattern is Section Type (e.g. AA, BB)
        section_type: int = zlib_data[zlib_start - 0x1]

        section_name: str = {0xAA: 'Firmware', 0xBB: 'Utilities'}.get(section_type, f'Unknown ({section_type:02X})')

        # Show extraction complete message for each main PFS ZLIB Section
        printer(message=f'Extracting Dell PFS {pfs_index} > {pfs_name} > {section_name}', padding=padding)

        # Set PFS ZLIB Section extraction subdirectory path
        section_path: str = os.path.join(extract_path, safe_name(in_name=section_name))

        # Create extraction subdirectory and delete old (if present, not in recursions)
        make_dirs(in_path=section_path, delete=not is_rec)

        # Store the compressed zlib stream start offset
        compressed_start: int = zlib_start + 0xB

        # Store the PFS ZLIB section header start offset
        header_start: int = zlib_start - 0x5

        # Store the PFS ZLIB section header contents (16 bytes)
        header_data: bytes = zlib_data[header_start:compressed_start]

        # Check if the PFS ZLIB section header Checksum XOR 8 is valid
        if checksum_8_xor(data=header_data[:0xF]) != header_data[0xF]:
            printer(message='Error: Invalid Dell PFS ZLIB section Header Checksum!', padding=padding)

            is_zlib_error = True

        # Store the compressed zlib stream size from the header contents
        compressed_size_hdr: int = int.from_bytes(header_data[:0x4], byteorder='little')

        # Store the compressed zlib stream end offset
        compressed_end: int = compressed_start + compressed_size_hdr

        # Store the compressed zlib stream contents
        compressed_data: bytes = zlib_data[compressed_start:compressed_end]

        # Check if the compressed zlib stream is complete, based on header
        if len(compressed_data) != compressed_size_hdr:
            printer(message='Error: Incomplete Dell PFS ZLIB section data (Header)!', padding=padding)

            is_zlib_error = True

        # Store the PFS ZLIB section footer contents (16 bytes)
        footer_data: bytes = zlib_data[compressed_end:compressed_end + 0x10]

        # Check if PFS ZLIB section footer was found in the section
        if not self._is_pfs_ftr(input_object=footer_data):
            printer(message='Error: This Dell PFS ZLIB section is corrupted!', padding=padding)

            is_zlib_error = True

        # Check if the PFS ZLIB section footer Checksum XOR 8 is valid
        if checksum_8_xor(data=footer_data[:0xF]) != footer_data[0xF]:
            printer(message='Error: Invalid Dell PFS ZLIB section Footer Checksum!', padding=padding)

            is_zlib_error = True

        # Store the compressed zlib stream size from the footer contents
        compressed_size_ftr: int = int.from_bytes(footer_data[:0x4], byteorder='little')

        # Check if the compressed zlib stream is complete, based on footer
        if compressed_size_ftr != compressed_size_hdr:
            printer(message='Error: Incomplete Dell PFS ZLIB section data (Footer)!', padding=padding)

            is_zlib_error = True

        # Decompress PFS ZLIB section payload
        try:
            if is_zlib_error:
                raise ValueError('ZLIB_ERROR_OCCURRED')  # ZLIB errors are critical

            section_data: bytes = zlib.decompress(compressed_data)  # ZLIB decompression
        except Exception as error:  # pylint: disable=broad-except
            printer(message=f'Error: Failed to decompress PFS ZLIB section: {error}!', padding=padding)

            section_data = zlib_data  # Fallback to raw ZLIB data upon critical error

        # Call the PFS Extract function on the decompressed PFS ZLIB Section
        self._pfs_extract(buffer=section_data, pfs_index=pfs_index, pfs_name=pfs_name, pfs_count=pfs_count,
                          extract_path=section_path, padding=padding)

    def _pfs_extract(self, buffer: bytes | bytearray, pfs_index: int, pfs_name: str, pfs_count: int,
                     extract_path: str, padding: int = 0) -> None:
        """ Parse & Extract Dell PFS Volume """

        # Show PFS Volume indicator
        if self.structure:
            printer(message='PFS Volume:', padding=padding)

        # Get PFS Header Structure values
        pfs_hdr: Any = ctypes_struct(buffer=buffer, start_offset=0, class_object=DellPfsHeader)

        # Validate that a PFS Header was parsed
        if pfs_hdr.Tag != b'PFS.HDR.':
            printer(message='Error: PFS Header could not be found!', padding=padding + 4)

            return  # Critical error, abort

        # Show PFS Header Structure info
        if self.structure:
            printer(message='PFS Header:\n', padding=padding + 4)

            pfs_hdr.struct_print(padding=padding + 8)

        # Validate that a known PFS Header Version was encountered
        self._chk_hdr_ver(version=pfs_hdr.HeaderVersion, text='PFS', padding=padding + 8)

        # Get PFS Payload Data
        pfs_payload: bytes = buffer[self.PFS_HEAD_LEN:self.PFS_HEAD_LEN + pfs_hdr.PayloadSize]

        # Parse all PFS Payload Entries/Components
        entry_index: int = 1  # Index number of each PFS Entry
        entry_start: int = 0  # Increasing PFS Entry starting offset
        entries_all: list[list] = []  # Storage for each PFS Entry details
        filename_info = b''  # Buffer for FileName Information Entry Data
        signature_info = b''  # Buffer for Signature Information Entry Data

        # Get PFS Entry Info
        pfs_entry_struct, pfs_entry_size = self._get_pfs_entry(buffer=pfs_payload, offset=entry_start)

        while len(pfs_payload[entry_start:entry_start + pfs_entry_size]) == pfs_entry_size:
            # Analyze PFS Entry Structure and get relevant info
            _, entry_version, entry_guid, entry_data, entry_data_sig, entry_met, entry_met_sig, next_entry = \
                self._parse_pfs_entry(entry_buffer=pfs_payload, entry_start=entry_start, entry_size=pfs_entry_size,
                                      entry_struct=pfs_entry_struct, text='PFS Entry', padding=padding)

            entry_type: str = 'OTHER'  # Adjusted later if PFS Entry is Zlib, PFAT, PFS Info, Model Info

            # Get PFS Information from the relevant (hardcoded) PFS Entry GUIDs
            if entry_guid in ['E0717CE3A9BB25824B9F0DC8FD041960', 'B033CB16EC9B45A14055F80E4D583FD3']:
                entry_type = 'NAME_INFO'

                filename_info = entry_data

            # Get Model Information from the relevant (hardcoded) PFS Entry GUID
            elif entry_guid == '6F1D619A22A6CB924FD4DA68233AE3FB':
                entry_type = 'MODEL_INFO'

            # Get Signature Information from the relevant (hardcoded) PFS Entry GUID
            elif entry_guid == 'D086AFEE3ADBAEA94D5CED583C880BB7':
                entry_type = 'SIG_INFO'

                signature_info = entry_data

            # Get Nested PFS from the relevant (hardcoded) PFS Entry GUID
            elif entry_guid == '900FAE60437F3AB14055F456AC9FDA84':
                entry_type = 'NESTED_PFS'  # Nested PFS are usually zlib-compressed so it might change to 'ZLIB' later

            # Store all relevant PFS Entry details
            entries_all.append([entry_index, entry_guid, entry_version, entry_type,
                                entry_data, entry_data_sig, entry_met, entry_met_sig])

            entry_index += 1  # Increase PFS Entry Index number for user-friendly output and name duplicates

            entry_start = next_entry  # Next PFS Entry starts after PFS Entry Metadata Signature

        # Parse all PFS Information Entries/Descriptors
        info_start: int = 0  # Increasing PFS Information Entry starting offset
        info_all: list[list] = []  # Storage for each PFS Information Entry details

        while len(filename_info[info_start:info_start + self.PFS_INFO_LEN]) == self.PFS_INFO_LEN:
            # Get PFS Information Header Structure info
            filename_info_hdr: Any = ctypes_struct(buffer=filename_info, start_offset=info_start,
                                                   class_object=DellPfsInfo)

            # Show PFS Information Header Structure info
            if self.structure:
                printer(message='PFS Filename Information Header:\n', padding=padding + 4)

                filename_info_hdr.struct_print(padding=padding + 8)

            # Validate that a known PFS Information Header Version was encountered
            if filename_info_hdr.HeaderVersion != 1:
                printer(message=f'Error: Unknown PFS Filename Information Header '
                                f'Version {filename_info_hdr.HeaderVersion}!', padding=padding + 8)

                break  # Skip PFS Information Entries/Descriptors in case of unknown PFS Information Header Version

            # Get PFS Information Header GUID in Big Endian format, in order
            # to match each Info to the equivalent stored PFS Entry details.
            entry_guid = f'{int.from_bytes(filename_info_hdr.GUID, byteorder="little"):0{0x10 * 2}X}'

            # Get PFS FileName Structure values
            entry_info_mod: Any = ctypes_struct(buffer=filename_info, start_offset=info_start + self.PFS_INFO_LEN,
                                                class_object=DellPfsName)

            # The PFS FileName Structure is not complete by itself. The size of the last field (Entry Name)
            # is determined from CharacterCount multiplied by 2 due to usage of UTF-16 2-byte Characters.
            # Any Entry Name leading and/or trailing space/null characters are stripped and Windows
            # reserved/illegal filename characters are replaced.
            name_start: int = info_start + self.PFS_INFO_LEN + self.PFS_NAME_LEN  # PFS Entry's FileName start offset

            name_size: int = entry_info_mod.CharacterCount * 2  # PFS Entry's FileName buffer total size

            name_data: bytes = filename_info[name_start:name_start + name_size]  # PFS Entry's FileName buffer

            # PFS Entry's FileName value
            entry_name: str = safe_name(in_name=name_data.decode('utf-16').strip())

            # Show PFS FileName Structure info
            if self.structure:
                printer(message='PFS FileName Entry:\n', padding=padding + 8)

                entry_info_mod.struct_print(name=entry_name, padding=padding + 12)

            # Get PFS FileName Version string via "Version" and "VersionType" fields
            # PFS FileName Version string must be preferred over PFS Entry's Version
            entry_version = self.get_entry_ver(version_fields=entry_info_mod.Version,
                                               version_types=entry_info_mod.VersionType)

            # Store all relevant PFS FileName details
            info_all.append([entry_guid, entry_name, entry_version])

            # The next PFS Information Header starts after the calculated FileName size
            # Two space/null characters seem to always exist after each FileName value
            info_start += (self.PFS_INFO_LEN + self.PFS_NAME_LEN + name_size + 0x2)

        # Parse Nested PFS Metadata when its PFS Information Entry is missing
        for entry in entries_all:
            _, entry_guid, _, entry_type, _, _, entry_metadata, _ = entry

            if entry_type == 'NESTED_PFS' and not filename_info:
                # When PFS Information Entry exists, Nested PFS Metadata contains only Model IDs
                # When it's missing, the Metadata structure is large and contains equivalent info
                if len(entry_metadata) >= self.PFS_META_LEN:
                    # Get Nested PFS Metadata Structure values
                    entry_info: Any = ctypes_struct(buffer=entry_metadata, start_offset=0, class_object=DellPfsMetadata)

                    # Show Nested PFS Metadata Structure info
                    if self.structure:
                        printer(message='PFS Metadata Information:\n', padding=padding + 4)

                        entry_info.struct_print(padding=padding + 8)

                    # As Nested PFS Entry Name, we'll use the actual PFS File Name
                    # Replace common Windows reserved/illegal filename characters
                    entry_name = safe_name(in_name=entry_info.FileName.decode('utf-8').removesuffix('.exe')
                                           .removesuffix('.bin'))

                    # As Nested PFS Entry Version, we'll use the actual PFS File Version
                    entry_version = entry_info.FileVersion.decode('utf-8')

                    # Store all relevant Nested PFS Metadata/Information details
                    info_all.append([entry_guid, entry_name, entry_version])

                    # Re-set Nested PFS Entry Version from Metadata
                    entry[2] = entry_version

        # Parse all PFS Signature Entries/Descriptors
        sign_start: int = 0  # Increasing PFS Signature Entry starting offset

        while len(signature_info[sign_start:sign_start + self.PFS_INFO_LEN]) == self.PFS_INFO_LEN:
            # Get PFS Information Header Structure info
            signature_info_hdr: Any = ctypes_struct(buffer=signature_info, start_offset=sign_start,
                                                    class_object=DellPfsInfo)

            # Show PFS Information Header Structure info
            if self.structure:
                printer(message='PFS Signature Information Header:\n', padding=padding + 4)

                signature_info_hdr.struct_print(padding=padding + 8)

            # Validate that a known PFS Information Header Version was encountered
            if signature_info_hdr.HeaderVersion != 1:
                printer(message=f'Error: Unknown PFS Signature Information Header '
                                f'Version {signature_info_hdr.HeaderVersion}!', padding=padding + 8)

                break  # Skip PFS Signature Entries/Descriptors in case of unknown Header Version

            # PFS Signature Entries have DellPfsInfo + DellPfsEntryR* + Sign Size [0x2] + Sign Data [Sig Size]
            pfs_entry_struct, pfs_entry_size = self._get_pfs_entry(buffer=signature_info,
                                                                   offset=sign_start + self.PFS_INFO_LEN)

            # Get PFS Entry Header Structure info
            entry_hdr: Any = ctypes_struct(buffer=signature_info, start_offset=sign_start + self.PFS_INFO_LEN,
                                           class_object=pfs_entry_struct)

            # Show PFS Information Header Structure info
            if self.structure:
                printer(message='PFS Information Entry:\n', padding=padding + 8)

                entry_hdr.struct_print(padding=padding + 12)

            # Show PFS Signature Size & Data (after DellPfsEntryR*)
            sign_info_start: int = sign_start + self.PFS_INFO_LEN + pfs_entry_size

            sign_size: int = int.from_bytes(signature_info[sign_info_start:sign_info_start + 0x2],
                                            byteorder='little')

            sign_data_raw: bytes = signature_info[sign_info_start + 0x2:sign_info_start + 0x2 + sign_size]

            sign_data_txt: str = f'{int.from_bytes(sign_data_raw, byteorder="little"):0{sign_size * 2}X}'

            if self.structure:
                printer(message='Signature Information:\n', padding=padding + 8)

                printer(message=f'Signature Size: 0x{sign_size:X}', padding=padding + 12, new_line=False)

                printer(message=f'Signature Data: {sign_data_txt[:32]} [...]', padding=padding + 12, new_line=False)

            # The next PFS Signature Entry/Descriptor starts after the previous Signature Data
            sign_start += (self.PFS_INFO_LEN + pfs_entry_size + 0x2 + sign_size)

        # Parse each PFS Entry Data for special types (zlib or PFAT)
        for entry in entries_all:
            _, _, _, entry_type, entry_data, _, _, _ = entry

            # Very small PFS Entry Data cannot be of special type
            if len(entry_data) < self.PFS_HEAD_LEN:
                continue

            # Check if PFS Entry contains zlib-compressed sub-PFS Volume
            pfs_zlib_offsets: list[int] = self._get_section_offsets(buffer=entry_data)

            # Parse PFS Entry which contains zlib-compressed sub-PFS Volume
            if pfs_zlib_offsets:
                entry_type = 'ZLIB'  # Re-set PFS Entry Type from OTHER to ZLIB, to use such info afterward

                pfs_count += 1  # Increase the count/index of parsed main PFS structures by one

                # Parse each sub-PFS ZLIB Section
                for offset in pfs_zlib_offsets:
                    # Get the Name of the zlib-compressed full PFS structure via the already stored PFS Information
                    # The zlib-compressed full PFS structure(s) are used to contain multiple FW (CombineBiosNameX)
                    # When zlib-compressed full PFS structure(s) exist within the main/first full PFS structure,
                    # its PFS Information should contain their names (CombineBiosNameX). Since the main/first
                    # full PFS structure has count/index 1, the rest start at 2+ and thus, their PFS Information
                    # names can be retrieved in order by subtracting 2 from the main/first PFS Information values
                    sub_pfs_name = f'{info_all[pfs_count - 2][1]} v{info_all[pfs_count - 2][2]}' \
                        if info_all else ' UNKNOWN'

                    # Set the sub-PFS output path (create sub-folders for each sub-PFS and its ZLIB sections)
                    sub_pfs_path: str = os.path.join(extract_path, f'{pfs_count} {safe_name(in_name=sub_pfs_name)}')

                    # Recursively call the PFS ZLIB Section Parser for the sub-PFS Volume (pfs_index = pfs_count)
                    self._pfs_section_parse(zlib_data=entry_data, zlib_start=offset, extract_path=sub_pfs_path,
                                            pfs_name=sub_pfs_name, pfs_index=pfs_count, pfs_count=pfs_count,
                                            is_rec=True, padding=padding + 4)

            # Initialize possible PFAT PFS Entry Offset
            pfat_pfs_header_off: int = 0

            while entry_data[pfat_pfs_header_off:pfat_pfs_header_off + 8] == b'PFS.HDR.':
                pfat_pfs_header: Any = ctypes_struct(buffer=entry_data, start_offset=pfat_pfs_header_off,
                                                     class_object=DellPfsHeader)

                # Show PFS Volume indicator
                if self.structure:
                    printer(message='PFS Volume:', padding=padding + 4)

                # Show sub-PFS Header Structure Info
                if self.structure:
                    printer(message='PFS Header:\n', padding=padding + 8)

                    pfat_pfs_header.struct_print(padding=padding + 12)

                # Validate that a known sub-PFS Header Version was encountered
                self._chk_hdr_ver(version=pfat_pfs_header.HeaderVersion, text='Sub-PFS', padding=padding + 12)

                pfat_pfs_entry_off: int = pfat_pfs_header_off + self.PFS_HEAD_LEN
                pfat_pfs_footer_off: int = pfat_pfs_entry_off + pfat_pfs_header.PayloadSize

                # Get sub-PFS Entry Structure and Size
                pfat_pfs_entry_struct, pfat_pfs_entry_size = self._get_pfs_entry(
                    buffer=entry_data, offset=pfat_pfs_entry_off)

                # Show sub-PFS Entry Structure Info
                _, _, _, _, _, _, _, _ = self._parse_pfs_entry(
                    entry_buffer=entry_data, entry_start=pfat_pfs_entry_off, entry_size=pfat_pfs_entry_size,
                    entry_struct=pfat_pfs_entry_struct, text='Sub-PFS Entry', padding=padding + 12)

                # Get sub-PFS Payload Data
                pfat_pfs_entry_payload: bytes = entry_data[pfat_pfs_entry_off:pfat_pfs_footer_off]

                # Get sub-PFS Footer Data
                pfat_pfs_entry_footer: bytes = entry_data[pfat_pfs_footer_off:pfat_pfs_footer_off + self.PFS_FOOT_LEN]

                # Next nested PFS or PFAT Entries start after PFS Header and Entry
                pfat_pfs_header_off += self.PFS_HEAD_LEN + pfat_pfs_entry_size

                if len(entry_data[pfat_pfs_header_off:]) <= self.PFAT_HDR_LEN:
                    continue

                pfat_intel_header: Any = ctypes_struct(buffer=entry_data, start_offset=pfat_pfs_header_off,
                                                       class_object=IntelBiosGuardHeader)

                # Parse PFS Entry which contains sub-PFS Volume with PFAT Payload
                if pfat_intel_header.get_platform_id().upper().startswith('DELL'):
                    # Re-set PFS Entry Type from OTHER to PFAT, to use such info afterward
                    entry_type = 'PFAT'

                    # Parse sub-PFS PFAT Volume
                    entry_data = self._parse_pfat_pfs(pfat_payload=pfat_pfs_entry_payload, padding=padding)

                    # Analyze sub-PFS Footer Structure
                self._chk_pfs_ftr(footer_buffer=pfat_pfs_entry_footer, data_buffer=pfat_pfs_entry_payload,
                                  data_size=pfat_pfs_header.PayloadSize, text='Sub-PFS', padding=padding + 4)

            # Adjust PFS Entry Data after parsing PFAT (same ZLIB raw data, not stored afterward)
            entry[4] = entry_data

            # Adjust PFS Entry Type from OTHER to PFAT or ZLIB (ZLIB is ignored at file extraction)
            entry[3] = entry_type

        # Name & Store each PFS Entry/Component Data, Data Signature, Metadata, Metadata Signature
        for entry in entries_all:
            file_index, file_guid, file_version, file_type, file_data, file_data_sig, file_meta, file_meta_sig = entry

            # Give Names to special PFS Entries, not covered by PFS Information
            if file_type == 'MODEL_INFO':
                file_name: str = 'Model Information'
            elif file_type == 'NAME_INFO':
                file_name = 'Filename Information'

                if not self.advanced:
                    continue  # Don't store Filename Information in non-advanced user mode
            elif file_type == 'SIG_INFO':
                file_name = 'Signature Information'

                if not self.advanced:
                    continue  # Don't store Signature Information in non-advanced user mode
            else:
                file_name = ''

            # Most PFS Entry Names & Versions are found at PFS Information via their GUID
            # Version can be found at DellPfsEntryR* but prefer PFS Information when possible
            for info in info_all:
                info_guid, info_name, info_version = info

                # Give proper Name & Version info if Entry/Information GUIDs match
                if info_guid == file_guid:
                    file_name = info_name

                    file_version = info_version

                    # PFS with zlib-compressed sub-PFS use the same GUID
                    info[0] = 'USED'

                    # Break at 1st Name match to not rename again from
                    # next zlib-compressed sub-PFS with the same GUID.
                    break

            # For both advanced & non-advanced users, the goal is to store final/usable files only
            # so empty or intermediate files such as sub-PFS, PFS w/ PFAT or zlib-PFS are skipped
            # Main/First PFS CombineBiosNameX Metadata files must be kept for accurate Model Information
            # All users should check these files in order to choose the correct CombineBiosNameX modules
            write_files: list[list] = []  # Initialize list of output PFS Entry files to be written/extracted

            is_zlib: bool = bool(file_type == 'ZLIB')  # Determine if PFS Entry Data was zlib-compressed

            if file_data and not is_zlib:
                write_files.append([file_data, 'data'])  # PFS Entry Data Payload

            if file_data_sig and self.advanced:
                write_files.append([file_data_sig, 'sign_data'])  # PFS Entry Data Signature

            if file_meta and (is_zlib or self.advanced):
                write_files.append([file_meta, 'meta'])  # PFS Entry Metadata Payload

            if file_meta_sig and self.advanced:
                write_files.append([file_meta_sig, 'sign_meta'])  # PFS Entry Metadata Signature

            # Write/Extract PFS Entry files
            for file in write_files:
                # Full PFS Entry Name
                full_name: str = f'{pfs_index} {pfs_name} -- {file_index} {file_name} v{file_version}'

                self._pfs_file_write(bin_buff=file[0], bin_name=file[1], bin_type=file_type, full_name=full_name,
                                     out_path=extract_path, padding=padding)

        # Get PFS Footer Data after PFS Header Payload
        pfs_footer: bytes = buffer[self.PFS_HEAD_LEN + pfs_hdr.PayloadSize:
                                   self.PFS_HEAD_LEN + pfs_hdr.PayloadSize + self.PFS_FOOT_LEN]

        # Analyze PFS Footer Structure
        self._chk_pfs_ftr(footer_buffer=pfs_footer, data_buffer=pfs_payload, data_size=pfs_hdr.PayloadSize,
                          text='PFS', padding=padding)

    def _parse_pfs_entry(self, entry_buffer: bytes | bytearray, entry_start: int, entry_size: int,
                         entry_struct: Any, text: str, padding: int = 0) -> tuple:
        """ Analyze Dell PFS Entry Structure """

        # Get PFS Entry Structure values
        pfs_entry: Any = ctypes_struct(buffer=entry_buffer, start_offset=entry_start, class_object=entry_struct)

        # Show PFS Entry Structure info
        if self.structure:
            printer(message='PFS Entry:\n', padding=padding + 4)

            pfs_entry.struct_print(padding=padding + 8)

        # Validate that a known PFS Entry Header Version was encountered
        self._chk_hdr_ver(version=pfs_entry.HeaderVersion, text=text, padding=padding + 8)

        # Validate that the PFS Entry Reserved field is empty
        if pfs_entry.Reserved != 0:
            printer(message=f'Error: Detected non-empty {text} Reserved field!', padding=padding + 8)

        # Get PFS Entry Version string via "Version" and "VersionType" fields
        entry_version: str = self.get_entry_ver(version_fields=pfs_entry.Version, version_types=pfs_entry.VersionType)

        # Get PFS Entry GUID in Big Endian format
        entry_guid: str = f'{int.from_bytes(pfs_entry.GUID, byteorder="little"):0{0x10 * 2}X}'

        # PFS Entry Data starts after the PFS Entry Structure
        entry_data_start: int = entry_start + entry_size
        entry_data_end: int = entry_data_start + pfs_entry.DataSize

        # PFS Entry Data Signature starts after PFS Entry Data
        entry_data_sig_start: int = entry_data_end
        entry_data_sig_end: int = entry_data_sig_start + pfs_entry.DataSigSize

        # PFS Entry Metadata starts after PFS Entry Data Signature
        entry_met_start: int = entry_data_sig_end
        entry_met_end: int = entry_met_start + pfs_entry.DataMetSize

        # PFS Entry Metadata Signature starts after PFS Entry Metadata
        entry_met_sig_start: int = entry_met_end
        entry_met_sig_end: int = entry_met_sig_start + pfs_entry.DataMetSigSize

        # Store PFS Entry Data
        entry_data: bytes = entry_buffer[entry_data_start:entry_data_end]

        # Store PFS Entry Data Signature
        entry_data_sig: bytes = entry_buffer[entry_data_sig_start:entry_data_sig_end]

        # Store PFS Entry Metadata
        entry_met: bytes = entry_buffer[entry_met_start:entry_met_end]

        # Store PFS Entry Metadata Signature
        entry_met_sig: bytes = entry_buffer[entry_met_sig_start:entry_met_sig_end]

        return (pfs_entry, entry_version, entry_guid, entry_data, entry_data_sig,
                entry_met, entry_met_sig, entry_met_sig_end)

    def _parse_pfat_pfs(self, pfat_payload: bytes | bytearray, padding: int = 0) -> bytes:
        """ Parse Dell PFS Volume with PFAT Payload """

        # Parse all sub-PFS Payload PFAT Entries
        pfat_entries_all: list[tuple] = []  # Storage for all sub-PFS PFAT Entries Order/Offset & Payload/Raw Data
        pfat_entry_start: int = 0  # Increasing sub-PFS PFAT Entry start offset
        pfat_entry_index: int = 1  # Increasing sub-PFS PFAT Entry count index

        # Get initial PFS PFAT Entry Size for loop
        _, pfs_entry_size = self._get_pfs_entry(buffer=pfat_payload, offset=0)

        # Initialize sub-PFS PFAT Signature length
        _pfat_sign_len: int = 0

        while len(pfat_payload[pfat_entry_start:pfat_entry_start + pfs_entry_size]) == pfs_entry_size:
            # Get sub-PFS PFAT Entry Structure & Size info
            pfat_entry_struct, pfat_entry_size = self._get_pfs_entry(buffer=pfat_payload, offset=pfat_entry_start)

            # Analyze sub-PFS PFAT Entry Structure and get relevant info
            pfat_entry, _, _, pfat_entry_data, _, pfat_entry_met, _, pfat_next_entry = \
                self._parse_pfs_entry(entry_buffer=pfat_payload, entry_start=pfat_entry_start,
                                      entry_size=pfat_entry_size, entry_struct=pfat_entry_struct,
                                      text='sub-PFS PFAT Entry', padding=padding + 4)

            # Each sub-PFS PFAT Entry includes an AMI BIOS Guard (a.k.a. PFAT) block at the beginning
            # We need to parse the PFAT block and remove its contents from the final Payload/Raw Data
            pfat_hdr_off: int = pfat_entry_start + pfat_entry_size  # PFAT block starts after PFS Entry

            # Get sub-PFS PFAT Header Structure values
            pfat_hdr: Any = ctypes_struct(buffer=pfat_payload, start_offset=pfat_hdr_off,
                                          class_object=IntelBiosGuardHeader)

            # Get ordinal value of the sub-PFS PFAT Entry Index
            pfat_entry_idx_ord: str = to_ordinal(in_number=pfat_entry_index)

            # Show sub-PFS PFAT Header Structure info
            if self.structure:
                printer(message=f'PFAT Block {pfat_entry_idx_ord} - Header:\n', padding=padding + 12)

                pfat_hdr.struct_print(padding=padding + 16)

            pfat_script_start: int = pfat_hdr_off + self.PFAT_HDR_LEN  # PFAT Block Script Start

            pfat_script_end: int = pfat_script_start + pfat_hdr.ScriptSize  # PFAT Block Script End

            pfat_script_data: bytes = pfat_payload[pfat_script_start:pfat_script_end]  # PFAT Block Script Data

            pfat_payload_start: int = pfat_script_end  # PFAT Block Payload Start (at Script end)

            pfat_payload_end: int = pfat_script_end + pfat_hdr.DataSize  # PFAT Block Data End

            pfat_payload_data: bytes = pfat_payload[pfat_payload_start:pfat_payload_end]  # PFAT Block Raw Data

            pfat_hdr_bgs_size: int = self.PFAT_HDR_LEN + pfat_hdr.ScriptSize  # PFAT Block Header & Script Size

            # The PFAT Script End should match the total Entry Data Size w/o PFAT block
            if pfat_hdr_bgs_size != pfat_entry.DataSize - pfat_hdr.DataSize:
                printer(message=f'Error: Detected sub-PFS PFAT Block {pfat_entry_idx_ord} Header & '
                                f'PFAT Size mismatch!', padding=padding + 16)

            # Get PFAT Header Flags (SFAM, ProtectEC, GFXMitDis, FTU, Reserved)
            is_sfam, _, _, _, _ = pfat_hdr.get_flags()

            ami_pfat_extract: AmiPfatExtract = AmiPfatExtract()

            # Parse sub-PFS PFAT Signature, if applicable (only when PFAT Header > SFAM flag is set)
            if is_sfam:
                if self.structure:
                    printer(message=f'PFAT Block {pfat_entry_idx_ord} - Signature:\n', padding=padding + 12)

                # Get sub-PFS PFAT Signature Size from Header pattern (not necessary for Dell PFS)
                if _pfat_sign_len == 0:
                    _pfat_sign_sig: bytes = pfat_hdr.get_hdr_marker()
                    _pfat_sign_pfs: int = pfs_entry_size + self.PFS_PFAT_LEN
                    _pfat_sign_max: int = _pfat_sign_pfs + ami_pfat_extract.PFAT_INT_SIG_MAX_LEN + len(_pfat_sign_sig)
                    _pfat_sign_lim: int = pfat_payload_end + _pfat_sign_max
                    _pfat_sign_off: int = pfat_payload.find(_pfat_sign_sig, pfat_payload_end, _pfat_sign_lim)
                    _pfat_sign_len = _pfat_sign_off - pfat_payload_end - _pfat_sign_pfs

                # Get sub-PFS PFAT Signature Structure values
                pfat_sign_len: int = ami_pfat_extract.parse_bg_sign(
                    input_data=pfat_payload, sign_offset=pfat_payload_end, sign_length=_pfat_sign_len,
                    print_info=self.structure, padding=padding + 16)

                if not len(pfat_payload[pfat_payload_end:pfat_payload_end + pfat_sign_len]
                           ) == pfat_sign_len == _pfat_sign_len:
                    printer(message=f'Error: Detected sub-PFS PFAT Block {pfat_entry_idx_ord} Signature '
                                    f'Size mismatch!', padding=padding + 12)

            # Show PFAT Script via BIOS Guard Script Tool
            if self.structure:
                printer(message=f'PFAT Block {pfat_entry_idx_ord} - Script:\n', padding=padding + 12)

                _ = ami_pfat_extract.parse_bg_script(script_data=pfat_script_data, padding=padding + 16)

            # The payload of sub-PFS PFAT Entries is not in proper order by default
            # We can get each payload's order from PFAT Script > OpCode #2 (set I0 imm)
            # PFAT Script OpCode #2 > Operand #3 stores the payload Offset in final image
            pfat_entry_off: int = int.from_bytes(pfat_script_data[0xC:0x10], byteorder='little')

            # We can get each payload's length from PFAT Script > OpCode #4 (set I2 imm)
            # PFAT Script OpCode #4 > Operand #3 stores the payload Length in final image
            pfat_entry_len: int = int.from_bytes(pfat_script_data[0x1C:0x20], byteorder='little')

            # Check that the PFAT Entry Length from Header & Script match
            if pfat_hdr.DataSize != pfat_entry_len:
                printer(message=f'Error: Detected sub-PFS PFAT Block {pfat_entry_idx_ord} Header & '
                                f'Script Size mismatch!', padding=padding + 12)

            # Initialize sub-PFS PFAT Entry Metadata Address
            pfat_entry_adr: int = pfat_entry_off

            # Parse sub-PFS PFAT Entry/Block Metadata
            if len(pfat_entry_met) >= self.PFS_PFAT_LEN:
                # Get sub-PFS PFAT Metadata Structure values
                pfat_met: Any = ctypes_struct(buffer=pfat_entry_met, start_offset=0, class_object=DellPfsPfatMetadata)

                # Store sub-PFS PFAT Entry Metadata Address
                pfat_entry_adr = pfat_met.Address

                # Show sub-PFS PFAT Metadata Structure info
                if self.structure:
                    printer(message=f'PFAT Block {pfat_entry_idx_ord} - Metadata:\n', padding=padding + 12)

                    pfat_met.struct_print(padding=padding + 16)

                # Another way to get each PFAT Entry Offset is from its Metadata, if applicable
                # Check that the PFAT Entry Offsets from PFAT Script and PFAT Metadata match
                if pfat_entry_off != pfat_met.Offset:
                    printer(message=f'Error: Detected sub-PFS PFAT Block {pfat_entry_idx_ord} Metadata & '
                                    f'PFAT Offset mismatch!', padding=padding + 16)

                    # Prefer Offset from Metadata, in case PFAT Script differs
                    pfat_entry_off = pfat_met.Offset

                # Another way to get each PFAT Entry Length is from its Metadata, if applicable
                # Check that the PFAT Entry Length from PFAT Script and PFAT Metadata match
                if not pfat_hdr.DataSize == pfat_entry_len == pfat_met.DataSize:
                    printer(message=f'Error: Detected sub-PFS PFAT Block {pfat_entry_idx_ord} Metadata & '
                                    f'PFAT Length mismatch!', padding=padding + 16)

                # Check that the PFAT Entry payload Size from PFAT Header matches the one from PFAT Metadata
                if pfat_hdr.DataSize != pfat_met.DataSize:
                    printer(message=f'Error: Detected sub-PFS PFAT Block {pfat_entry_idx_ord} Metadata & '
                                    f'PFAT Block Size mismatch!', padding=padding + 16)

            # Get sub-PFS Entry Raw Data by subtracting PFAT Header & Script from PFAT Entry Data
            pfat_entry_data_raw: bytes = pfat_entry_data[pfat_hdr_bgs_size:]

            # The sub-PFS Entry Raw Data (w/o PFAT Header & Script) should match with the PFAT Block payload
            if pfat_entry_data_raw != pfat_payload_data:
                printer(message=f'Error: Detected sub-PFS PFAT Block {pfat_entry_idx_ord} w/o PFAT & '
                                f'PFAT Block Data mismatch!', padding=padding + 16)

                # Prefer Data from PFAT Block, in case PFAT Entry differs
                pfat_entry_data_raw = pfat_payload_data

            # Store each sub-PFS PFAT Entry/Block Offset, Address, Ordinal Index and Payload/Raw Data
            # Goal is to sort these based on Offset first and Address second, in cases of same Offset
            # For example, Precision 3430 has two PFAT Entries with the same Offset of 0x40000 at both
            # BG Script and PFAT Metadata but their PFAT Metadata Address is 0xFF040000 and 0xFFA40000
            pfat_entries_all.append((pfat_entry_off, pfat_entry_adr, pfat_entry_idx_ord, pfat_entry_data_raw))

            # Check if next sub-PFS PFAT Entry offset is valid
            if pfat_next_entry <= 0:
                printer(message=f'Error: Detected sub-PFS PFAT Block {pfat_entry_idx_ord} with invalid '
                                f'next PFAT Block offset!', padding=padding + 16)

                # Avoid a potential infinite loop if next sub-PFS PFAT Entry offset is bad
                pfat_next_entry += pfs_entry_size

            # Next sub-PFS PFAT Entry starts after sub-PFS Entry Metadata Signature
            pfat_entry_start = pfat_next_entry

            pfat_entry_index += 1

        # Sort all sub-PFS PFAT Entries based on their Offset/Address
        pfat_entries_all.sort()

        block_start_exp: int = 0  # Initialize sub-PFS PFAT Entry expected Offset
        total_pfat_data: bytes = b''  # Initialize final/ordered sub-PFS Entry Data

        # Parse all sorted sub-PFS PFAT Entries and merge their payload/data
        for block_start, _, block_index, block_data in pfat_entries_all:
            # Fill any data gaps between sorted sub-PFS PFAT Entries with padding
            # For example, Precision 7960 v0.16.68 has gap at 0x1190000-0x11A0000
            block_data_gap: int = block_start - block_start_exp

            if block_data_gap > 0:
                if self.structure:
                    printer(message=f'Warning: Filled sub-PFS PFAT {block_index} data gap 0x{block_data_gap:X} '
                                    f'[0x{block_start_exp:X}-0x{block_start:X}]!', padding=padding + 8)

                # Use 0xFF padding to fill in data gaps in PFAT UEFI firmware images
                total_pfat_data += b'\xFF' * block_data_gap

            total_pfat_data += block_data  # Append sorted sub-PFS PFAT Entry payload/data

            block_start_exp = len(total_pfat_data)  # Set next sub-PFS PFAT Entry expected Start

        # Verify that the end offset of the last PFAT Entry matches the final sub-PFS Entry Data Size
        if len(total_pfat_data) != pfat_entries_all[-1][0] + len(pfat_entries_all[-1][3]):
            printer(message='Error: Detected sub-PFS PFAT total buffer size and '
                            'last block end mismatch!', padding=padding + 8)

        return total_pfat_data

    @staticmethod
    def _get_pfs_entry(buffer: bytes | bytearray, offset: int) -> tuple:
        """ Get Dell PFS Entry Structure & Size via its Version """

        # PFS Entry Version
        pfs_entry_ver: int = int.from_bytes(buffer[offset + 0x10:offset + 0x14], byteorder='little')

        if pfs_entry_ver == 1:
            return DellPfsEntryR1, ctypes.sizeof(DellPfsEntryR1)

        if pfs_entry_ver == 2:
            return DellPfsEntryR2, ctypes.sizeof(DellPfsEntryR2)

        return DellPfsEntryR2, ctypes.sizeof(DellPfsEntryR2)

    @staticmethod
    def get_entry_ver(version_fields: bytes, version_types: bytes) -> str:
        """ Determine Dell PFS Entry Version string """

        version: str = ''  # Initialize Version string

        # Version Type (1 byte) determines the type of Version Value (2 bytes)
        # Version Type 'N' is Number, 'A' is Text and ' ' is Empty/Unused
        for index, field in enumerate(version_fields):
            eol: str = '' if index == len(version_fields) - 1 else '.'

            if version_types[index] == 65:
                version += f'{field:X}{eol}'  # 0x41 = ASCII
            elif version_types[index] == 78:
                version += f'{field:d}{eol}'  # 0x4E = Number
            elif version_types[index] in (0, 32):
                version = version.strip('.')  # 0x00 or 0x20 = Unused
            else:
                version += f'{field:X}{eol}'  # Unknown

        return version

    @staticmethod
    def _chk_hdr_ver(version: int, text: str, padding: int = 0) -> None:
        """ Check if Dell PFS Header Version is known """

        if version not in (1, 2):
            printer(message=f'Error: Unknown {text} Header Version {version}!', padding=padding)

    def _chk_pfs_ftr(self, footer_buffer: bytes | bytearray, data_buffer: bytes | bytearray,
                     data_size: int, text: str, padding: int = 0) -> None:
        """ Analyze Dell PFS Footer Structure """

        # Get PFS Footer Structure values
        pfs_ftr: Any = ctypes_struct(buffer=footer_buffer, start_offset=0, class_object=DellPfsFooter)

        # Validate that a PFS Footer was parsed
        if pfs_ftr.Tag == b'PFS.FTR.':
            # Show PFS Footer Structure info
            if self.structure:
                printer(message='PFS Footer:\n', padding=padding + 4)

                pfs_ftr.struct_print(padding=padding + 8)
        else:
            printer(message=f'Error: {text} Footer could not be found!', padding=padding + 4)

        # Validate that PFS Header Payload Size matches the one at PFS Footer
        if data_size != pfs_ftr.PayloadSize:
            printer(message=f'Error: {text} Header & Footer Payload Size mismatch!', padding=padding + 4)

        # Calculate the PFS Payload Data CRC-32 w/ Vector 0
        pfs_ftr_crc: int = ~zlib.crc32(data_buffer, 0) & 0xFFFFFFFF

        # Validate PFS Payload Data Checksum via PFS Footer
        if pfs_ftr.Checksum != pfs_ftr_crc:
            printer(message=f'Error: Invalid {text} Footer Payload Checksum!', padding=padding + 4)

    def _pfs_file_write(self, bin_buff: bytes | bytearray, bin_name: str, bin_type: str,
                        full_name: str, out_path: str, padding: int = 0) -> None:
        """ Write/Extract Dell PFS Entry Files (Data, Metadata, Signature) """

        # Store Data/Metadata Signature (advanced users only)
        if bin_name.startswith('sign'):
            final_name: str = f'{safe_name(in_name=full_name)}.{bin_name.split("_")[1]}.sig'

            final_path: str = os.path.join(out_path, final_name)

            with open(final_path, 'wb') as pfs_out:
                pfs_out.write(bin_buff)  # Write final Data/Metadata Signature

            return  # Skip further processing for Signatures

        # Store Data/Metadata Payload (simpler Data/Metadata Extension for non-advanced users)
        bin_ext: str = f'.{bin_name}.bin' if self.advanced else '.bin'

        # Some Data may be Text or XML files with useful information for non-advanced users
        final_data, file_ext = self._bin_is_text(buffer=bin_buff, file_type=bin_type,
                                                 is_metadata=bin_name == 'meta', padding=padding)

        is_text: bool = isinstance(final_data, str)

        final_name = f'{safe_name(in_name=full_name)}{bin_ext[:-4] + file_ext if is_text else bin_ext}'

        final_path = os.path.join(out_path, final_name)

        # Write final Data/Metadata Payload
        with open(final_path, 'w' if is_text else 'wb', encoding='utf-8' if is_text else None) as pfs_out:
            pfs_out.write(final_data)

    def _bin_is_text(self, buffer: bytes | bytearray, file_type: str, is_metadata: bool, padding: int = 0) -> tuple:
        """
        Check if Dell PFS Entry file/data is Text/XML and Convert

        Only for non-advanced users due to signature (.sig) invalidation
        """

        extension: str = '.bin'

        if self.advanced:
            return buffer, extension

        buffer_text: str = ''

        if b',END' in buffer[-0x8:]:  # Text Type 1
            extension = '.txt'

            buffer_text = buffer.decode('utf-8').split(',END')[0].replace(';', '\n')
        elif buffer.startswith(b'VendorName=Dell'):  # Text Type 2
            extension = '.txt'

            buffer_text = buffer.split(b'\x00')[0].decode('utf-8').replace(';', '\n')
        elif b'<Rimm x-schema="' in buffer[:0x50]:  # XML Type
            extension = '.xml'

            buffer_text = buffer.decode('utf-8')
        elif file_type in ('NESTED_PFS', 'ZLIB') and is_metadata and len(buffer) == self.PFS_META_LEN:  # Text Type 3
            extension = '.txt'

            with io.StringIO() as text_buffer, contextlib.redirect_stdout(text_buffer):
                ctypes_struct(buffer=buffer, start_offset=0, class_object=DellPfsMetadata).struct_print(padding=0)

                # noinspection PyUnresolvedReferences
                buffer_text = text_buffer.getvalue()

        # Show Model/PCR XML Information, if applicable
        # Metadata is shown at initial DellPfsMetadata analysis
        if self.structure and buffer_text and not is_metadata:
            metadata_info_type: str = {".txt": "Model", ".xml": "PCR XML"}[extension]

            printer(message=f'PFS {metadata_info_type} Information:\n', padding=padding + 8)

            for line in [line for line in buffer_text.split('\n') if line]:
                printer(message=line.strip('\r'), padding=padding + 12, new_line=False)

        return buffer_text or buffer, extension



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""


# Get Checksum 16-bit
def checksum_16(data: bytes | bytearray, value: int = 0, order: str = 'little') -> int:
    """ Calculate Checksum-16 of data, controlling IV and Endianess """

    for idx in range(0, len(data), 2):
        # noinspection PyTypeChecker
        value += int.from_bytes(data[idx:idx + 2], byteorder=order)  # type: ignore

    value &= 0xFFFF

    return value


# Get Checksum 8-bit XOR
def checksum_8_xor(data: bytes | bytearray, value: int = 0) -> int:
    """ Calculate Checksum-8 XOR of data, controlling IV """

    for byte in data:
        value ^= byte

    value ^= 0x0

    return value



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Award BIOS Extract
Award BIOS Module Extractor
Copyright (C) 2018-2024 Plato Mavropoulos
"""

import os

from biosutilities.common.compression import szip_decompress
from biosutilities.common.paths import (clear_readonly, delete_file, extract_folder, is_file_read,
                                        make_dirs, safe_name)
from biosutilities.common.patterns import PAT_AWARD_LZH
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility


class AwardBiosExtract(BIOSUtility):
    """ Award BIOS Module Extractor """

    TITLE: str = 'Award BIOS Module Extractor'

    def check_format(self) -> bool:
        """ Check if input is Award BIOS image """

        return bool(PAT_AWARD_LZH.search(self.input_buffer))

    def parse_format(self) -> bool:
        """ Parse & Extract Award BIOS image """

        make_dirs(in_path=self.extract_path)

        for lzh_match in PAT_AWARD_LZH.finditer(self.input_buffer):
            lzh_type: str = lzh_match.group(0).decode('utf-8')

            lzh_text: str = f'LZH-{lzh_type.strip("-").upper()}'

            lzh_bgn: int = lzh_match.start()

            mod_bgn: int = lzh_bgn - 0x2
            hdr_len: int = self.input_buffer[mod_bgn]
            mod_len: int = int.from_bytes(self.input_buffer[mod_bgn + 0x7:mod_bgn + 0xB], byteorder='little')
            mod_end: int = lzh_bgn + hdr_len + mod_len

            mod_bin: bytes = self.input_buffer[mod_bgn:mod_end]

            if len(mod_bin) != 0x2 + hdr_len + mod_len:
                printer(message=f'Error: Skipped incomplete LZH stream at 0x{mod_bgn:X}!',
                        padding=self.padding, new_line=True)

                continue

            if len(mod_bin) >= 0x16:
                tag_txt: str = safe_name(in_name=mod_bin[0x16:0x16 + mod_bin[0x15]].decode('utf-8', 'ignore').strip())
            else:
                tag_txt = f'{mod_bgn:X}_{mod_end:X}'

            printer(message=f'{lzh_text} > {tag_txt} [0x{mod_bgn:06X}-0x{mod_end:06X}]', padding=self.padding)

            mod_path: str = os.path.join(self.extract_path, tag_txt)

            lzh_path: str = f'{mod_path}.lzh'

            with open(lzh_path, 'wb') as lzh_file:
                lzh_file.write(mod_bin)  # Store LZH archive

            # 7-Zip returns critical exit code (i.e. 2) if LZH CRC is wrong, do not check result
            szip_decompress(in_path=lzh_path, out_path=self.extract_path, in_name=lzh_text,
                            padding=self.padding + 4)

            # Manually check if 7-Zip extracted LZH due to its CRC check issue
            if is_file_read(in_path=mod_path):
                clear_readonly(in_path=lzh_path)

                delete_file(in_path=lzh_path)  # Successful extraction, delete LZH archive

                award_bios_extract: AwardBiosExtract = AwardBiosExtract(
                    input_object=mod_path, extract_path=extract_folder(mod_path), padding=self.padding + 8)

                # Extract any nested LZH archives
                if award_bios_extract.check_format():
                    award_bios_extract.parse_format()

        return True



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Phoenix TDK Extract
Phoenix TDK Packer Extractor
Copyright (C) 2021-2024 Plato Mavropoulos
"""

import ctypes
import logging
import lzma
import os

from re import Match
from typing import Any, Final

from pefile import PE

from biosutilities.common.paths import is_file_read, make_dirs, safe_name
from biosutilities.common.executables import ms_pe, ms_pe_info
from biosutilities.common.patterns import PAT_MICROSOFT_MZ, PAT_MICROSOFT_PE, PAT_PHOENIX_TDK
from biosutilities.common.structs import CHAR, ctypes_struct, UINT32
from biosutilities.common.system import printer
from biosutilities.common.templates import BIOSUtility


class PhoenixTdkHeader(ctypes.LittleEndianStructure):
    """ Phoenix TDK Header """

    _pack_ = 1
    _fields_ = [
        ('Tag',         CHAR * 8),      # 0x00
        ('Size',        UINT32),        # 0x08
        ('Count',       UINT32)         # 0x0C
        # 0x10
    ]

    def _get_tag(self) -> str:
        return self.Tag.decode('utf-8', 'ignore').strip()

    def struct_print(self, padding: int = 0) -> None:
        """ Display structure information """

        printer(message=['Tag    :', self._get_tag()], padding=padding, new_line=False)
        printer(message=['Size   :', f'0x{self.Size:X}'], padding=padding, new_line=False)
        printer(message=['Entries:', self.Count], padding=padding, new_line=False)


class PhoenixTdkEntry(ctypes.LittleEndianStructure):
    """ Phoenix TDK Entry """

    _pack_ = 1
    _fields_ = [
        ('Name',        CHAR * 256),    # 0x000
        ('Offset',      UINT32),        # 0x100
        ('Size',        UINT32),        # 0x104
        ('Compressed',  UINT32),        # 0x108
        ('Reserved',    UINT32)         # 0x10C
        # 0x110
    ]

    COMP: Final[dict[int, str]] = {0: 'None', 1: 'LZMA'}

    def __init__(self, mz_base: int, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.mz_base: int = mz_base

    def get_name(self) -> str:
        """ Get TDK Entry decoded name """

        return self.Name.decode('utf-8', 'replace').strip()

    def get_offset(self) -> int:
        """ Get TDK Entry absolute offset """

        return self.mz_base + self.Offset

    def get_compression(self) -> str:
        """ Get TDK Entry compression type """

        return self.COMP.get(self.Compressed, f'Unknown ({self.Compressed})')

    def struct_print(self, padding: int) -> None:
        """ Display structure information """

        printer(message=['Name       :', self.get_name()], padding=padding, new_line=False)
        printer(message=['Offset     :', f'0x{self.get_offset():X}'], padding=padding, new_line=False)
        printer(message=['Size       :', f'0x{self.Size:X}'], padding=padding, new_line=False)
        printer(message=['Compression:', self.get_compression()], padding=padding, new_line=False)
        printer(message=['Reserved   :', f'0x{self.Reserved:X}'], padding=padding, new_line=False)


class PhoenixTdkExtract(BIOSUtility):
    """ Phoenix TDK Packer Extractor """

    TITLE: str = 'Phoenix TDK Packer Extractor'

    TDK_HDR_LEN: Final[int] = ctypes.sizeof(PhoenixTdkHeader)
    TDK_MOD_LEN: Final[int] = ctypes.sizeof(PhoenixTdkEntry)

    TDK_DUMMY_LEN: Final[int] = 0x200

    def check_format(self) -> bool:
        """ Check if input contains valid Phoenix TDK image """

        return bool(self._get_phoenix_tdk(in_buffer=self.input_buffer)[1] is not None)

    def parse_format(self) -> bool:
        """ Parse & Extract Phoenix Tools Development Kit (TDK) Packer """

        exit_code: int = 0

        make_dirs(in_path=self.extract_path)

        printer(message='Phoenix Tools Development Kit Packer', padding=self.padding)

        base_off, pack_off = self._get_phoenix_tdk(in_buffer=self.input_buffer)

        # Parse TDK Header structure
        tdk_hdr: Any = ctypes_struct(buffer=self.input_buffer, start_offset=pack_off, class_object=PhoenixTdkHeader)

        # Print TDK Header structure info
        printer(message='Phoenix TDK Header:\n', padding=self.padding + 4)

        tdk_hdr.struct_print(padding=self.padding + 8)

        # Check if reported TDK Header Size matches manual TDK Entry Count calculation
        if tdk_hdr.Size != self.TDK_HDR_LEN + self.TDK_DUMMY_LEN + tdk_hdr.Count * self.TDK_MOD_LEN:
            printer(message='Error: Phoenix TDK Header Size & Entry Count mismatch!\n', padding=self.padding + 8)

            exit_code = 1

        # Store TDK Entries offset after the placeholder data
        entries_off: int = pack_off + self.TDK_HDR_LEN + self.TDK_DUMMY_LEN

        # Parse and extract each TDK Header Entry
        for entry_index in range(tdk_hdr.Count):
            # Parse TDK Entry structure
            tdk_mod: Any = ctypes_struct(
                buffer=self.input_buffer, start_offset=entries_off + entry_index * self.TDK_MOD_LEN,
                class_object=PhoenixTdkEntry, param_list=[base_off]
            )

            # Print TDK Entry structure info
            printer(message=f'Phoenix TDK Entry ({entry_index + 1}/{tdk_hdr.Count}):\n', padding=self.padding + 8)

            tdk_mod.struct_print(padding=self.padding + 12)

            # Get TDK Entry raw data Offset (TDK Base + Entry Offset)
            mod_off: int = tdk_mod.get_offset()

            # Check if TDK Entry raw data Offset is valid
            if mod_off >= len(self.input_buffer):
                printer(message='Error: Phoenix TDK Entry > Offset is out of bounds!\n', padding=self.padding + 12)

                exit_code = 2

            # Store TDK Entry raw data (relative to TDK Base, not TDK Header)
            mod_data: bytes = self.input_buffer[mod_off:mod_off + tdk_mod.Size]

            # Check if TDK Entry raw data is complete
            if len(mod_data) != tdk_mod.Size:
                printer(message='Error: Phoenix TDK Entry > Data is truncated!\n', padding=self.padding + 12)

                exit_code = 3

            # Check if TDK Entry Reserved is present
            if tdk_mod.Reserved:
                printer(message='Error: Phoenix TDK Entry > Reserved is not empty!\n', padding=self.padding + 12)

                exit_code = 4

            # Decompress TDK Entry raw data, when applicable (i.e. LZMA)
            if tdk_mod.get_compression() == 'LZMA':
                try:
                    mod_data = lzma.LZMADecompressor().decompress(data=mod_data)
                except Exception as error:  # pylint: disable=broad-except
                    printer(message=f'Error: Phoenix TDK Entry > LZMA decompression failed: {error}!\n',
                            padding=self.padding + 12)

                    exit_code = 5

            # Generate TDK Entry file name, avoid crash if Entry data is bad
            mod_name: str = tdk_mod.get_name() or f'Unknown_{entry_index + 1:02d}.bin'

            # Generate TDK Entry file data output path
            mod_file: str = os.path.join(self.extract_path, safe_name(mod_name))

            # Account for potential duplicate file names
            if is_file_read(in_path=mod_file):
                mod_file += f'_{entry_index + 1:02d}'

            # Save TDK Entry data to output file
            with open(mod_file, 'wb') as out_file:
                out_file.write(mod_data)

        return exit_code == 0

    @staticmethod
    def _get_tdk_base(in_buffer: bytes | bytearray, pack_off: int) -> int | None:
        """ Get Phoenix TDK Executable (MZ) Base Offset """

        # Initialize Phoenix TDK Base MZ Offset
        tdk_base_off: int | None = None

        # Scan input file for all Microsoft executable patterns (MZ) before TDK Header Offset
        mz_all: list[Match[bytes]] = [mz for mz in PAT_MICROSOFT_MZ.finditer(in_buffer) if mz.start() < pack_off]

        # Phoenix TDK Header structure is an index table for all TDK files
        # Each TDK file is referenced from the TDK Packer executable base
        # The TDK Header is always at the end of the TDK Packer executable
        # Thus, prefer the TDK Packer executable (MZ) closest to TDK Header
        # For speed, check MZ closest to (or at) 0x0 first (expected input)
        mz_ord: list[Match[bytes]] = [mz_all[0]] + list(reversed(mz_all[1:]))

        # Parse each detected MZ
        for mz_match in mz_ord:
            mz_off: int = mz_match.start()

            # MZ (DOS) > PE (NT) image Offset is found at offset 0x3C-0x40 relative to MZ base
            pe_off: int = mz_off + int.from_bytes(in_buffer[mz_off + 0x3C:mz_off + 0x40], byteorder='little')

            # Skip MZ (DOS) with bad PE (NT) image Offset
            if pe_off == mz_off or pe_off >= pack_off:
                continue

            # Check if potential MZ > PE image magic value is valid
            if PAT_MICROSOFT_PE.search(in_buffer[pe_off:pe_off + 0x4]):
                try:
                    # Parse detected MZ > PE > Image, quickly (fast_load)
                    pe_file: PE | None = ms_pe(in_file=in_buffer[mz_off:], silent=True)

                    if not pe_file:
                        raise RuntimeError('Failed to parse detected MZ > PE > Image!')

                    # Parse detected MZ > PE > Info
                    pe_info: dict = ms_pe_info(pe_file=pe_file, silent=True)

                    # Parse detected MZ > PE > Info > Product Name
                    pe_name: bytes = pe_info.get(b'ProductName', b'')
                except Exception as error:  # pylint: disable=broad-except
                    # Any error means no MZ > PE > Info > Product Name
                    logging.debug('Error: Invalid potential MZ > PE match at 0x%X: %s', pe_off, error)

                    pe_name = b''

                # Check for valid Phoenix TDK Packer PE > Product Name
                # Expected value is "TDK Packer (Extractor for Windows)"
                if pe_name.upper().startswith(b'TDK PACKER'):
                    # Set TDK Base Offset to valid TDK Packer MZ offset
                    tdk_base_off = mz_off

            # Stop parsing detected MZ once TDK Base Offset is found
            if tdk_base_off is not None:
                break
        else:
            # No TDK Base Offset could be found, assume 0x0
            tdk_base_off = 0x0

        return tdk_base_off

    def _get_phoenix_tdk(self, in_buffer: bytes | bytearray) -> tuple:
        """ Scan input buffer for valid Phoenix TDK image """

        # Scan input buffer for Phoenix TDK pattern
        tdk_match: Match[bytes] | None = PAT_PHOENIX_TDK.search(in_buffer)

        if not tdk_match:
            return None, None

        # Set Phoenix TDK Header ($PACK) Offset
        tdk_pack_off: int = tdk_match.start()

        # Get Phoenix TDK Executable (MZ) Base Offset
        tdk_base_off: int | None = self._get_tdk_base(in_buffer, tdk_pack_off)

        return tdk_base_off, tdk_pack_off



#!/usr/bin/env python3 -B
# coding=utf-8

"""
Copyright (C) 2022-2024 Plato Mavropoulos
"""

import pefile

from biosutilities.common.system import printer
from biosutilities.common.texts import file_to_bytes


def is_ms_pe(in_file: str | bytes) -> bool:
    """ Check if input is a PE file """

    return bool(ms_pe(in_file=in_file, silent=True))


def ms_pe(in_file: str | bytes, padding: int = 0, fast: bool = True, silent: bool = False) -> pefile.PE | None:
    """ Get pefile object from PE file """

    pe_file: pefile.PE | None = None

    try:
        # Analyze detected MZ > PE image buffer
        pe_file = pefile.PE(data=file_to_bytes(in_file), fast_load=fast)
    except Exception as error:  # pylint: disable=broad-except
        if not silent:
            filename: str = in_file if isinstance(in_file, str) else 'buffer'

            printer(message=f'Error: Could not get pefile object from {filename}: {error}!', padding=padding)

    return pe_file


def ms_pe_desc(pe_file: pefile.PE, padding: int = 0, silent: bool = False) -> bytes:
    """ Get PE description from pefile object info """

    return ms_pe_info(pe_file=pe_file, padding=padding, silent=silent).get(b'FileDescription', b'')


def ms_pe_info(pe_file: pefile.PE, padding: int = 0, silent: bool = False) -> dict:
    """ Get PE info from pefile object """

    pe_info: dict = {}

    try:
        # When fast_load is used, IMAGE_DIRECTORY_ENTRY_RESOURCE must be parsed prior to FileInfo > StringTable
        pe_file.parse_data_directories(directories=[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_RESOURCE']])

        # Retrieve MZ > PE > FileInfo > StringTable information
        pe_info = pe_file.FileInfo[0][0].StringTable[0].entries
    except Exception as error:  # pylint: disable=broad-except
        if not silent:
            printer(message=f'Error: Could not get PE info from pefile object: {error}!', padding=padding)

    return pe_info


def ms_pe_info_show(pe_file: pefile.PE, padding: int = 0) -> None:
    """ Print PE info from pefile StringTable """

    pe_info: dict = ms_pe_info(pe_file=pe_file, padding=padding)

    if isinstance(pe_info, dict):
        for title, value in pe_info.items():
            info_title: str = title.decode('utf-8', 'ignore').strip()
            info_value: str = value.decode('utf-8', 'ignore').strip()

            if info_title and info_value:
                printer(message=f'{info_title}: {info_value}', padding=padding, new_line=False)
